"""
Exceptions relating to local operations.
"""
from musify.exception import MusifyError


class LocalError(MusifyError):
    """
    Exception raised for local errors.

    :param message: Explanation of the error.
    """
    def __init__(self, message: str | None = None):
        self.message = message
        super().__init__(message)


class LocalItemError(LocalError):
    """
    Exception raised for local item errors.

    :param message: Explanation of the error.
    :param kind: The item type related to the error.
    """
    def __init__(self, message: str | None = None, kind: str | None = None):
        self.message = message
        self.kind = kind
        formatted = f"{kind} | {message}" if kind else message
        super().__init__(formatted)


class LocalCollectionError(LocalError):
    """
    Exception raised for local collection errors.

    :param message: Explanation of the error.
    :param kind: The collection type related to the error.
    """
    def __init__(self, message: str | None = None, kind: str | None = None):
        self.message = message
        self.kind = kind
        formatted = f"{kind} | {message}" if kind else message
        super().__init__(formatted)


###########################################################################
## Track errors
###########################################################################
class TagError(LocalError):
    """Exception raised for errors related to track tag errors."""


###########################################################################
## Library errors
###########################################################################
class LocalLibraryError(LocalError):
    """Exception raised for errors related to :py:class:`LocalLibrary` logic."""


class MusicBeeError(LocalLibraryError):
    """Exception raised for errors related to :py:class:`MusicBee` logic."""


class MusicBeeIDError(MusicBeeError):
    """Exception raised for errors related to MusicBee IDs."""


class XMLReaderError(MusicBeeError):
    """Exception raised for errors related to reading a MusicBee library XML file."""



"""
Core abstract classes for the :py:mod:`Local` module.
"""
from abc import ABCMeta

from musify.base import MusifyItemSettable
from musify.file.base import File


class LocalItem(File, MusifyItemSettable, metaclass=ABCMeta):
    """Generic base class for locally-stored items"""

    __slots__ = ()
    __attributes_classes__ = (File, MusifyItemSettable)



"""
Implements a :py:class:`RemoteLibrary` for Spotify.
"""
from collections.abc import Collection, Iterable
from typing import Any

from musify.libraries.remote.core.library import RemoteLibrary
from musify.libraries.remote.core.types import RemoteObjectType
from musify.libraries.remote.spotify.api import SpotifyAPI
from musify.libraries.remote.spotify.factory import SpotifyObjectFactory
from musify.libraries.remote.spotify.object import SpotifyTrack, SpotifyAlbum, SpotifyArtist, SpotifyPlaylist


class SpotifyLibrary(RemoteLibrary[SpotifyAPI, SpotifyPlaylist, SpotifyTrack, SpotifyAlbum, SpotifyArtist]):
    """
    Represents a Spotify library. Provides various methods for manipulating
    tracks and playlists across an entire Spotify library collection.
    """

    __slots__ = ()
    __attributes_classes__ = RemoteLibrary

    @staticmethod
    def _validate_item_type(items: Any | Iterable[Any]) -> bool:
        if isinstance(items, Iterable):
            return all(isinstance(item, SpotifyTrack) for item in items)
        return isinstance(items, SpotifyTrack)

    def _create_factory(self, api: SpotifyAPI) -> SpotifyObjectFactory:
        return SpotifyObjectFactory(api=api)

    def _filter_playlists(self, responses: list[dict[str, Any]]) -> list[dict[str, Any]]:
        pl_total = len(responses)
        pl_names_filtered = self.playlist_filter({response["name"] for response in responses})
        responses = [response for response in responses if response["name"] in pl_names_filtered]

        self.logger.debug(
            f"Filtered out {pl_total - len(responses)} playlists from "
            f"{pl_total} {self.api.source} available playlists"
            if (pl_total - len(responses)) > 0 else f"{len(responses)} playlists found"
        )

        return responses

    def _get_total_tracks(self, responses: list[dict[str, Any]]) -> int:
        return sum(pl["tracks"]["total"] for pl in responses)

    async def enrich_tracks(
            self, features: bool = False, analysis: bool = False, albums: bool = False, artists: bool = False
    ) -> None:
        """
        Enriches the ``features``, ``analysis``, ``albums``, and/or ``artists`` data for currently loaded tracks.

        :param features: Load all audio features (e.g. BPM, tempo, key etc.)
        :param analysis: Load all audio analyses (technical audio data).
            WARNING: can be very slow as calls to this endpoint cannot be batched i.e. one call per track.
        :param albums: Reload albums for all tracks, adding extra album data e.g. genres, popularity
        :param artists: Reload artists for all tracks, adding extra artist data e.g. genres, popularity, followers
        """
        if not self.tracks or not any((features, analysis, albums, artists)):
            return
        self.logger.debug(f"Enrich {self.api.source} tracks: START")
        self.logger.info(
            f"\33[1;95m  >\33[1;97m Enriching {len(self.tracks)} {self.api.source} tracks \33[0m"
        )

        tracks = [track for track in self.tracks if track.has_uri]
        await self.api.extend_tracks(tracks, features=features, analysis=analysis)

        # enrich on list of URIs to avoid duplicate calls for same items
        if albums:
            album_uris: set[str] = {track.response["album"]["uri"] for track in self.tracks}
            album_responses = await self.api.get_items(album_uris, kind=RemoteObjectType.ALBUM, extend=False)
            for album in album_responses:
                album.pop("tracks", None)

            albums = {response["uri"]: response for response in album_responses}
            for track in self.tracks:
                track.response["album"] = albums[track.response["album"]["uri"]]

        if artists:
            artist_uris: set[str] = {artist["uri"] for track in self.tracks for artist in track.response["artists"]}
            artist_responses = await self.api.get_items(artist_uris, kind=RemoteObjectType.ARTIST, extend=False)

            artists = {response["uri"]: response for response in artist_responses}
            for track in self.tracks:
                track.response["artists"] = [artists[artist["uri"]] for artist in track.response["artists"]]

        for track in self.tracks:
            track.refresh(skip_checks=False)  # tracks are popped from albums so checks should skip by default anyway

        self.logger.debug(f"Enrich {self.api.source} tracks: DONE\n")

    async def enrich_saved_albums(self) -> None:
        """Extends the tracks data for currently loaded albums, getting all available tracks data for each album"""
        if not self.albums or all(len(album) == album.track_total for album in self.albums):
            return
        self.logger.debug(f"Enrich {self.api.source} artists: START")
        self.logger.info(
            f"\33[1;95m  >\33[1;97m Enriching {len(self.albums)} {self.api.source} albums \33[0m"
        )

        kind = RemoteObjectType.ALBUM
        key = self.api.collection_item_map[kind]

        await self.logger.get_asynchronous_iterator(
            [self.api.extend_items(album, kind=kind, key=key) for album in self.albums],
            desc="Getting saved album tracks",
            unit="albums"
        )

        for album in self.albums:
            album.refresh(skip_checks=False)

            for track in album.tracks:  # add tracks from this album to the user's saved tracks
                if track not in self.tracks:
                    self._tracks.append(track)

        self.logger.debug(f"Enrich {self.api.source} artists: DONE\n")

    async def enrich_saved_artists(self, tracks: bool = False, types: Collection[str] = ()) -> None:
        """
        Gets all albums for current loaded following artists.

        :param tracks: When True, also get all tracks for each album.
        :param types: Provide a list of albums types to get to limit the types of albums loaded.
            Select from ``{"album", "single", "compilation", "appears_on"}``.
        """
        if not self.artists:
            return
        self.logger.debug(f"Enrich {self.api.source} artists: START")
        self.logger.info(
            f"\33[1;95m  >\33[1;97m Enriching {len(self.artists)} {self.api.source} artists \33[0m"
        )

        await self.api.get_artist_albums(self.artists, types=types)

        if tracks:
            kind = RemoteObjectType.ALBUM
            key = self.api.collection_item_map[kind]

            responses_albums = [album for artist in self.artists for album in artist.albums]
            await self.logger.get_asynchronous_iterator(
                [self.api.extend_items(album, kind=kind, key=key) for album in responses_albums],
                desc="Getting album tracks",
                unit="albums"
            )
            for album in responses_albums:
                album.refresh(skip_checks=False)

        self.logger.debug(f"Enrich {self.api.source} artists: DONE\n")



"""
Meta-functions for providing reports to the user based on comparisons between objects implemented in this package.
"""
import logging
from collections.abc import Iterable

from musify.base import MusifyItem
from musify.field import TagField, Fields, ALL_FIELDS, TagFields
from musify.libraries.core.collection import MusifyCollection
from musify.libraries.core.object import Library, Playlist
from musify.libraries.local.library import LocalLibrary
from musify.logger import MusifyLogger
from musify.logger import REPORT
from musify.types import UnitIterable
from musify.utils import align_string, get_max_width, to_collection


def report_playlist_differences(
        source: Library | Iterable[Playlist], reference: Library | Iterable[Playlist]
) -> dict[str, dict[str, tuple[MusifyItem, ...]]]:
    """
    Generate a report on the differences between two library's playlists.

    :param source: A source library object containing the source playlists, or a collection of playlists .
    :param reference: A comparative library object or collection of playlists
        containing the playlists to compare to the source's playlists.
    :return: Report on extra, missing, and unavailable tracks for the reference library.
    """
    # noinspection PyTypeChecker
    logger: MusifyLogger = logging.getLogger(__name__)
    logger.debug("Report library differences: START")

    extra: dict[str, tuple[MusifyItem, ...]] = {}
    missing: dict[str, tuple[MusifyItem, ...]] = {}
    unavailable: dict[str, tuple[MusifyItem, ...]] = {}

    log_message = "Reporting on playlist differences"
    if isinstance(source, Library) and isinstance(reference, Library):
        log_message += f" {source.source} and {reference.source} libraries"

    if isinstance(source, Library):
        source_name = source.source
        source = source.playlists
    else:
        source_name = "source"
        source = {pl.name: pl for pl in source}

    if isinstance(reference, Library):
        reference_name = reference.source
        reference = reference.playlists
    else:
        reference_name = "reference"
        reference = {pl.name: pl for pl in reference}

    max_width = get_max_width(source.keys())
    logger.info(f"\33[1;95m ->\33[1;97m {log_message} \33[0m")
    logger.print_line()

    for name, pl_source in source.items():
        pl_reference = reference.get(name, [])

        # get differences
        source_no_uri = tuple(item for item in pl_source if not item.has_uri)
        source_extra = tuple(item for item in pl_source if item not in pl_reference)
        reference_no_uri = tuple(item for item in pl_reference if not item.has_uri)
        reference_extra = tuple(item for item in pl_reference if item not in pl_source)

        extra[name] = reference_extra
        missing[name] = source_extra
        unavailable[name] = source_no_uri + reference_no_uri

        logger.report(
            f"\33[97m{align_string(name, max_width=max_width)} \33[0m|"
            f"\33[92m{len(reference_extra):>6} extra \33[0m|"
            f"\33[91m{len(source_extra):>6} missing \33[0m|"
            f"\33[93m{len(source_no_uri) + len(reference_no_uri):>6} unavailable \33[0m|"
            f"\33[94m{f"{len(pl_source):>6} in {source_name}"} \33[0m|"
            f"\33[96m{f"{len(pl_reference):>6} in {reference_name}"} \33[0m"
        )

    report: dict[str, dict[str, tuple[MusifyItem, ...]]] = {
        "Source ✗ | Compare ✓": extra,
        "Source ✓ | Compare ✗": missing,
        "Items unavailable (no URI)": unavailable
    }

    logger.report(
        f"\33[1;96m{'TOTALS':<{max_width}} \33[0m|"
        f"\33[1;92m{sum(map(len, extra.values())):>6} extra \33[0m|"
        f"\33[1;91m{sum(map(len, missing.values())):>6} missing \33[0m|"
        f"\33[1;93m{sum(map(len, unavailable.values())):>6} unavailable \33[0m|"
        f"\33[1;94m{len(source):>6} {"playlists":<{len(f"in {source_name}")}} \33[0m|"
        f"\33[1;96m{len(reference):>6} {"playlists":<{len(f"in {reference_name}")}} \33[0m"
    )
    logger.print_line(REPORT)
    logger.debug("Report library differences: DONE\n")
    return report


def report_missing_tags(
        collections: LocalLibrary | Iterable[MusifyCollection],
        tags: UnitIterable[TagField] = TagFields.ALL,
        match_all: bool = False
) -> dict[str, dict[MusifyItem, tuple[str, ...]]]:
    """
    Generate a report on the items with a set of collections that have missing tags.

    :param collections: A collection of item collections to report on.
        If a local library is given, use the albums of the library as the collections to report on.
    :param tags: List of tags to consider missing.
    :param match_all: When True, item counts as missing tags if item is missing ``all`` of the given tags.
        When False, item counts as missing tags when missing only one of the given tags.
    :return: Report on collections by name which have items with missing tags.
    """
    # noinspection PyTypeChecker
    logger: MusifyLogger = logging.getLogger(__name__)
    logger.debug("Report missing tags: START")

    if isinstance(collections, LocalLibrary):
        collections = collections.albums

    item_total = sum(map(len, collections))
    tag_names = _get_tag_names(logger=logger, tags=tags, item_total=item_total, match_all=match_all)

    missing: dict[str, dict[MusifyItem, tuple[str, ...]]] = {}
    for collection in collections:
        missing_collection: dict[MusifyItem, tuple[str, ...]] = {}
        for item in collection.items:
            missing_tags: list[str] = [tag for tag in tag_names if item[tag] is None]
            if "has_uri" in missing_tags:
                missing_tags[missing_tags.index("has_uri")] = Fields.URI.name.lower()
            if "has_image" in missing_tags:
                missing_tags[missing_tags.index("has_image")] = Fields.IMAGES.name.lower()
            if all(tag in missing_tags for tag in tag_names) if match_all else missing_tags:
                missing_collection[item] = tuple(missing_tags)

        if missing_collection:
            missing[collection.name] = missing_collection

    if not missing:
        logger.debug("Report missing tags: DONE\n")
        return missing

    _log_missing_tags(logger=logger, missing=missing)

    missing_tags_all = {tag for items in missing.values() for tags in items.values() for tag in tags}
    tag_order = [field.name.lower() for field in ALL_FIELDS]
    logger.info(
        f"    \33[94mFound {sum(map(len, missing.values()))} items with "
        f"{'all' if match_all else 'any'} missing tags\33[0m: \n"
        f"    \33[90m{', '.join(sorted(missing_tags_all, key=lambda x: tag_order.index(x)))}\33[0m"
    )
    logger.print_line()
    logger.debug("Report missing tags: DONE\n")
    return missing


def _get_tag_names(logger: MusifyLogger, tags: UnitIterable[TagField], item_total: int, match_all: bool) -> list[str]:
    tags = to_collection(tags, set)
    tag_order = [field.name.lower() for field in ALL_FIELDS]
    # noinspection PyTypeChecker
    tag_names_set = set(TagField.__tags__) if Fields.ALL in tags else TagField.to_tags(tags)
    tag_names: list[str] = list(sorted(tag_names_set, key=lambda x: tag_order.index(x)))

    logger.info(
        f"\33[1;95m ->\33[1;97m "
        f"Checking {item_total} items for {'all' if match_all else 'any'} missing tags: \n"
        f"    \33[90m{', '.join(tag_names)}\33[0m"
    )

    # switch out tag names with expected attribute names to check on
    if Fields.URI in tags or Fields.ALL in tags:
        tag_names[tag_names.index(Fields.URI.name.lower())] = "has_uri"
    if Fields.IMAGES in tags or Fields.ALL in tags:
        tag_names[tag_names.index(Fields.IMAGES.name.lower())] = "has_image"

    return tag_names


def _log_missing_tags(logger: MusifyLogger, missing: dict[str, dict[MusifyItem, tuple[str, ...]]]) -> None:
    all_keys = {item.name for items in missing.values() for item in items}
    max_width = get_max_width(all_keys)

    logger.print_line(REPORT)
    logger.report("\33[1;94mFound the following missing items by collection: \33[0m")
    logger.print_line(REPORT)

    for name, result in missing.items():
        logger.report(f"\33[1;91m -> {name} \33[0m")
        for item, tags in result.items():
            n = align_string(item.name, max_width=max_width)
            logger.report(f"\33[96m{n} \33[0m| \33[93m{', '.join(tags)} \33[0m")
        logger.print_line(REPORT)



"""
Generic processors that can be used in any module in the packages.

Examples:
    * Filtering down a collection items based on some predefined rules
    * Sorting a collection of items based on some predefined rules
    * Matching a collection of objects to some predefined rules
    * Limiting a collection of items based on some predefined rules
"""



"""
Base classes for all processors in this module. Also contains decorators for use in implementations.
"""
import logging
from abc import ABCMeta, abstractmethod
from collections.abc import Mapping, Callable, Collection, Iterable, MutableSequence
from functools import partial, update_wrapper
from typing import Any, Optional

from musify.logger import MusifyLogger
from musify.printer import PrettyPrinter
from musify.processors.exception import ProcessorLookupError
from musify.utils import get_user_input, get_max_width, align_string


class Processor(PrettyPrinter, metaclass=ABCMeta):
    """Generic base class for processors"""

    __slots__ = ()


class InputProcessor(Processor, metaclass=ABCMeta):
    """
    Processor that gets user input as part of it processing.

    Contains methods for getting user input and printing formatted options text to the terminal.
    """

    __slots__ = ("logger",)

    def __init__(self):
        # noinspection PyTypeChecker
        #: The :py:class:`MusifyLogger` for this  object
        self.logger: MusifyLogger = logging.getLogger(__name__)

    def _get_user_input(self, text: str | None = None) -> str:
        """Print dialog with optional text and get the user's input."""
        inp = get_user_input(text)
        self.logger.debug(f"User input: {inp}")
        return inp.strip()

    @staticmethod
    def _format_help_text(options: Mapping[str, str], header: MutableSequence[str] | None = None) -> str:
        """Format help text with a given mapping of options. Add an option header to include before options."""
        max_width = get_max_width(options)

        help_text = header or []
        help_text.append("\n\t\33[96mEnter one of the following: \33[0m\n\t")
        help_text.extend(
            f"\33[1m{align_string(k, max_width=max_width)}\33[0m{': ' + v or ''}" for k, v in options.items()
        )

        return "\n\t".join(help_text) + '\n'


# noinspection PyPep8Naming,SpellCheckingInspection
class dynamicprocessormethod:
    """
    Decorator for methods on a class decorated with the :py:func:`processor` decorator

    This assigns the method a processor method which can be dynamically called by the processor class.
    Optionally, provide a list of alternative names from which this processor method can also be called.
    """

    def __new__(cls, *args, **__):
        func: Optional[Callable] = next((a for a in args if callable(a)), None)
        self = partial(cls, *args) if func is None else super().__new__(cls)
        return update_wrapper(self, func)

    def __init__(self, *args: str | Callable):
        self.func = next((a for a in args if callable(a)), None)
        self.alternative_names = tuple(a for a in args if isinstance(a, str))
        self.instance_ = None

    def __get__(self, instance, owner):
        self.instance_ = instance
        return self.__call__

    def __call__(self, *args, **kwargs) -> Any:
        return self.func(self.instance_, *args, **kwargs) if self.instance_ else self.func(*args, **kwargs)


# noinspection SpellCheckingInspection
class DynamicProcessor(Processor, metaclass=ABCMeta):
    """
    Base class for implementations with :py:func:`dynamicprocessormethod` methods.

    Classes that implement this base class have a ``__processormethods__`` class attribute
    which is a list of strings of all the processor methods this class contains.
    If a :py:func:`dynamicprocessormethod` has alternative method names, these names will be added
    to the class' ``__dict__`` as callable methods which point to the decorated method.

    Optionally, you may also define a ``_processor_method_fmt`` class method which
    applies some transformation to all method names.
    The transformed method name is then appended to the class' ``__dict__``.
    The transformation is always applied before extending the class with any given
    alternative method names.
    """

    __slots__ = ("_processor_name",)

    #: The set of processor methods on this processor and any alternative names for them.
    __processormethods__: frozenset[str] = frozenset()

    @property
    def processor_methods(self) -> frozenset[str]:
        """String representation of all available processor names of this object"""
        return frozenset(self._processor_method_fmt(name) for name in self.__processormethods__)

    @classmethod
    def _processor_method_fmt(cls, name: str) -> str:
        """A custom formatter to apply to the dynamic processor name"""
        return name

    def __new__(cls, *_, **__):
        processor_methods = list(cls.__processormethods__)

        for method in cls.__dict__.copy().values():
            if not isinstance(method, dynamicprocessormethod):
                continue

            processor_methods.append(method.__name__)
            transformed_name = cls._processor_method_fmt(method.__name__)
            if transformed_name != method.__name__:
                processor_methods.append(transformed_name)
                setattr(cls, transformed_name, method)

            processor_methods.extend(method.alternative_names)
            for name in method.alternative_names:
                setattr(cls, cls._processor_method_fmt(name), method)

        cls.__processormethods__ = frozenset(processor_methods)
        return super().__new__(cls)

    def __init__(self):
        self._processor_name: str | None = None

    def _set_processor_name(self, value: str | None, fail_on_empty: bool = True):
        """Verifies and sets the condition name"""
        formatted_processor_names = map(lambda x: x.lstrip("_"), self.processor_methods)

        if value is None:
            if not fail_on_empty:
                self._processor_name = None
                return
            raise ProcessorLookupError(f"No condition given. Choose from: {", ".join(formatted_processor_names)}")

        name = self._processor_method_fmt(value)
        if name not in self.processor_methods:
            raise ProcessorLookupError(
                f"{value!r} condition is not valid. Choose from: {", ".join(formatted_processor_names)}"
            )

        self._processor_name = name

    @property
    def _processor_method(self) -> Callable:
        """The callable method of the dynamic processor"""
        return getattr(self, self._processor_name)

    def __call__(self, *args, **kwargs) -> Any:
        """Run the dynamic processor"""
        return self._processor_method(*args, **kwargs)


class Filter[T](Processor, metaclass=ABCMeta):
    """Base class for filtering down values based on some settings"""

    __slots__ = ("_transform",)

    @property
    @abstractmethod
    def ready(self) -> bool:
        """Does this filter have valid settings and can process values"""
        raise NotImplementedError

    @abstractmethod
    def process(self, values: Collection[T], *args, **kwargs) -> Collection[T]:
        """Apply this filter's settings to the given values"""
        raise NotImplementedError

    @property
    def transform(self) -> Callable[[Any], Any]:
        """
        Transform the input ``value`` to the value that should be used when comparing against this filter's settings
        Simply returns the given ``value`` at baseline unless overridden.
        """
        return self._transform

    @transform.setter
    def transform(self, transformer: Callable[[Any], Any]):
        self._transform = transformer

    def __init__(self):
        self._transform = lambda x: x

    def __call__(self, *args, **kwargs) -> Collection[T]:
        return self.process(*args, **kwargs)

    def __bool__(self):
        return self.ready


class FilterComposite[T](Filter[T], Collection[Filter], metaclass=ABCMeta):
    """Composite filter which filters based on many :py:class:`Filter` objects"""

    __slots__ = ("filters",)

    @property
    def ready(self):
        return any(filter_.ready for filter_ in self.filters)

    @property
    def transform(self) -> Callable[[Any], Any]:
        return lambda _: None

    @transform.setter
    def transform(self, transformer: Callable[[Any], Any]):
        for filter_ in self.filters:
            filter_.transform = transformer

    def __init__(self, *filters: Filter[T], **__):
        super().__init__()

        #: The filters to use when processing
        self.filters = filters

    def __iter__(self):
        def flat_filter_list(filter_: Filter | Collection[Filter]) -> Iterable[Filter]:
            """
            Get flat list of all :py:class:`Filter` objects in the given Filter,
            flattening out any :py:class:`FilterComposite` objects
            """
            if isinstance(filter_, FilterComposite):
                return iter(filter_)
            return [filter_]
        return (f for filter_ in self.filters for f in flat_filter_list(filter_))

    def __len__(self):
        return len(self.filters)

    def __contains__(self, item: Any):
        return item in self.filters



"""
Configuration relating to the :py:mod:`Remote` module.

This configuration can be used to inject dependencies into dependencies throughout the module.
"""
import inspect
from dataclasses import dataclass
from functools import partial

from musify.libraries.remote.core.api import RemoteAPI
from musify.libraries.remote.core.base import RemoteObject
from musify.libraries.remote.core.object import RemoteTrack, RemoteAlbum, RemotePlaylist, RemoteArtist
from musify.libraries.remote.core.types import RemoteObjectType


@dataclass
class RemoteObjectFactory[A: RemoteAPI, PL: RemotePlaylist, TR: RemoteTrack, AL: RemoteAlbum, AR: RemoteArtist]:
    """Stores the key object classes for a remote source"""
    #: The playlist type for this remote source
    playlist: type[PL]
    #: The track type for this remote source
    track: type[TR]
    #: The album type for this remote source
    album: type[AL]
    #: The artist type for this remote source
    artist: type[AR]
    #: An optional :py:class:`RemoteAPI` object to pass to each object on instantiation
    api: A = None

    def __getitem__(self, __key: RemoteObjectType) -> type[RemoteObject]:
        return self.__getattribute__(__key.name.lower())

    def __getattribute__(self, __name: str):
        attribute = object.__getattribute__(self, __name)
        if inspect.isclass(attribute) and issubclass(attribute, RemoteObject) and self.api is not None:
            executable = partial(attribute, api=self.api)

            # need to assign the class methods back to the partial object to ensure near seamless user use
            for key in dir(attribute):
                value = getattr(attribute, key)
                if inspect.ismethod(value) and not key.startswith("_"):
                    setattr(executable, key, partial(value, api=self.api))
            return executable
        return attribute



"""
The core Field enum for a :py:class:`LocalTrack` representing all possible tags/metadata/properties.
"""
from musify.field import TagFields, Fields
from musify.field import TrackFieldMixin


class LocalTrackField(TrackFieldMixin):
    """Represents all currently supported fields for objects of type :py:class:`LocalTrack`"""
    ALL = TagFields.ALL.value

    TITLE = TagFields.TITLE.value
    ARTIST = TagFields.ARTIST.value
    ARTISTS = Fields.ARTISTS.value
    ALBUM = TagFields.ALBUM.value
    ALBUM_ARTIST = TagFields.ALBUM_ARTIST.value
    TRACK = TagFields.TRACK_NUMBER.value + 500
    TRACK_NUMBER = TagFields.TRACK_NUMBER.value
    TRACK_TOTAL = TagFields.TRACK_TOTAL.value
    GENRES = TagFields.GENRES.value
    DATE = TagFields.DATE.value
    YEAR = TagFields.YEAR.value
    MONTH = TagFields.MONTH.value
    DAY = TagFields.DAY.value
    BPM = TagFields.BPM.value
    KEY = TagFields.KEY.value
    DISC = TagFields.DISC_NUMBER.value + 500
    DISC_NUMBER = TagFields.DISC_NUMBER.value
    DISC_TOTAL = TagFields.DISC_TOTAL.value
    COMPILATION = TagFields.COMPILATION.value
    COMMENTS = TagFields.COMMENTS.value
    IMAGES = TagFields.IMAGES.value
    LENGTH = TagFields.LENGTH.value
    RATING = TagFields.RATING.value

    # file properties
    PATH = TagFields.PATH.value
    FOLDER = TagFields.FOLDER.value
    FILENAME = TagFields.FILENAME.value
    EXT = TagFields.EXT.value
    SIZE = TagFields.SIZE.value
    TYPE = TagFields.TYPE.value
    CHANNELS = TagFields.CHANNELS.value
    BIT_RATE = TagFields.BIT_RATE.value
    BIT_DEPTH = TagFields.BIT_DEPTH.value
    SAMPLE_RATE = TagFields.SAMPLE_RATE.value

    # date properties
    DATE_MODIFIED = TagFields.DATE_MODIFIED.value
    DATE_ADDED = TagFields.DATE_ADDED.value
    LAST_PLAYED = TagFields.LAST_PLAYED.value

    # miscellaneous properties
    PLAY_COUNT = TagFields.PLAY_COUNT.value

    # remote properties
    URI = TagFields.URI.value



"""
The XAutoPF implementation of a :py:class:`LocalPlaylist`.
"""
from collections.abc import Collection, Mapping, Sequence
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Self

from musify.base import MusifyItem, Result
from musify.exception import FieldError, MusifyImportError
from musify.field import Fields, Field, TagFields
from musify.file.base import File
from musify.file.path_mapper import PathMapper
from musify.libraries.local.playlist.base import LocalPlaylist
from musify.libraries.local.track import LocalTrack
from musify.libraries.remote.core.wrangle import RemoteDataWrangler
from musify.printer import PrettyPrinter
from musify.processors.compare import Comparer
from musify.processors.exception import SorterProcessorError
from musify.processors.filter import FilterDefinedList, FilterComparers
from musify.processors.filter_matcher import FilterMatcher
from musify.processors.limit import ItemLimiter, LimitType
from musify.processors.sort import ItemSorter, ShuffleMode
from musify.utils import to_collection

try:
    import xmltodict
except ImportError:
    xmltodict = None

REQUIRED_MODULES = [xmltodict]

AutoMatcher = FilterMatcher[
    LocalTrack, FilterDefinedList[LocalTrack], FilterDefinedList[LocalTrack], FilterComparers[LocalTrack]
]


@dataclass(frozen=True)
class SyncResultXAutoPF(Result):
    """Stores the results of a sync with a local XAutoPF playlist."""
    #: The total number of tracks in the playlist before the sync.
    start: int
    #: The number of tracks that matched the include settings before the sync.
    start_included: int
    #: The number of tracks that matched the exclude settings before the sync.
    start_excluded: int
    #: The number of tracks that matched all the :py:class:`Comparer` settings before the sync.
    start_compared: int
    #: Was a limiter present on the playlist before the sync.
    start_limiter: bool
    #: Was a sorter present on the playlist before the sync.
    start_sorter: bool

    #: The total number of tracks in the playlist after the sync.
    final: int
    #: The number of tracks that matched the include settings after the sync.
    final_included: int
    #: The number of tracks that matched the exclude settings after the sync.
    final_excluded: int
    #: The number of tracks that matched all the :py:class:`Comparer` settings after the sync.
    final_compared: int
    #: Was a limiter present on the playlist after the sync.
    final_limiter: bool
    #: Was a sorter present on the playlist after the sync.
    final_sorter: bool


class XAutoPF(LocalPlaylist[AutoMatcher]):
    """
    For reading and writing data from MusicBee's auto-playlist format.

    :param path: Absolute path of the playlist.
        If the playlist ``path`` given does not exist, a new playlist will be created on :py:meth:`save`
    :param path_mapper: Optionally, provide a :py:class:`PathMapper` for paths stored in the playlist file.
        Useful if the playlist file contains relative paths and/or paths for other systems that need to be
        mapped to absolute, system-specific paths to be loaded and back again when saved.
    :param remote_wrangler: Not used by this object.
    """

    __slots__ = ("_parser", "_limiter_deduplication")
    __attributes_ignore__ = ("limiter_deduplication",)

    valid_extensions = frozenset({".xautopf"})

    #: The initial values to use as the base XML when creating a new XAutoPF file from scratch.
    #: Certain settings in the 'Source' key relating to processors that this program recognises
    #: will be set on initialisation. Hence, any default values assigned to these keys will be overridden.
    default_xml = {
        "SmartPlaylist": {
            "@SaveStaticCopy": "False",
            "@LiveUpdating": "True",
            "@Layout": "4",
            "@LayoutGroupBy": "0",
            "@ConsolidateAlbums": "False",
            "@MusicLibraryPath": "",
            "Source": {"@Type": "1", "Description": None}
        }
    }

    @property
    def description(self):
        return self._parser.description

    @description.setter
    def description(self, value: str | None):
        self._parser.description = value

    @property
    def limiter_deduplication(self) -> bool:
        """This setting controls whether duplicates should be filtered out before running limiter operations."""
        return self._limiter_deduplication

    @limiter_deduplication.setter
    def limiter_deduplication(self, value: bool):
        self._limiter_deduplication = value

    @property
    def image_links(self):
        return {}

    def __init__(
            self,
            path: str | Path,
            path_mapper: PathMapper = PathMapper(),
            remote_wrangler: RemoteDataWrangler = None
    ):
        if xmltodict is None:
            raise MusifyImportError(f"Cannot create {self.__class__.__name__} object. Required modules: xmltodict")

        super().__init__(path=path, path_mapper=path_mapper, remote_wrangler=remote_wrangler)

        self._parser: XMLPlaylistParser | None = None
        self._limiter_deduplication: bool = False

    async def load(self, tracks: Collection[LocalTrack] = ()) -> Self:
        """
        Read the playlist file and update the tracks in this playlist instance.

        :param tracks: Available Tracks to search through for matches.
            If no tracks are given, the playlist will be loaded empty.
        :return: Self
        """
        self._parser = XMLPlaylistParser(path=self.path, path_mapper=self.path_mapper)
        if self.path.is_file():
            await self._parser.load()
        else:  # this is a new playlist, assign default values to parser
            self._parser.xml = deepcopy(self.default_xml)
            self._parser.parse_matcher()
            self._parser.parse_limiter()
            self._parser.parse_sorter()

        self.matcher = self._parser.get_matcher()
        self.limiter = self._parser.get_limiter()
        self.sorter = self._parser.get_sorter()
        self._limiter_deduplication = self._parser.limiter_deduplication

        tracks_list = list(tracks)
        self.sorter.sort_by_field(tracks_list, field=Fields.LAST_PLAYED, reverse=True)

        self._match(tracks=tracks, reference=tracks_list[0] if len(tracks) > 0 else None)
        self._limit(ignore=self.matcher.exclude)
        self._sort()

        self._original = self.tracks.copy()

        return self

    def _limit(self, ignore: Collection[LocalTrack]) -> None:
        if self.limiter is not None and self.tracks is not None and self.limiter_deduplication:
            # preprocess tracks by applying deduplication first before sending to the actual limiter
            tracks_keys_seen = set()
            tracks_deduplicated: list[LocalTrack] = []

            for track in self.tracks:
                track_key = "_".join([track.title, track.artist])
                if track in ignore or track_key not in tracks_keys_seen:
                    tracks_keys_seen.add(track_key)
                    tracks_deduplicated.append(track)

            self.tracks = tracks_deduplicated
        super()._limit(ignore=ignore)

    async def save(self, dry_run: bool = True, *_, **__) -> SyncResultXAutoPF:
        """
        Write the tracks in this Playlist and its settings (if applicable) to file.

        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The results of the sync as a :py:class:`SyncResultXAutoPF` object.
        """
        initial = deepcopy(self._parser)
        initial_count = len(self._original)

        parser = self._parser if not dry_run else deepcopy(self._parser)
        parser.parse_matcher(self.matcher)
        parser.parse_exception_paths(self.matcher, items=self.tracks, original=self._original)
        parser.parse_limiter(self.limiter, deduplicate=self.limiter_deduplication)
        parser.parse_sorter(self.sorter)
        await parser.save(dry_run=dry_run)

        if not dry_run:
            self._original = self.tracks.copy()

        def _get_paths_sum(psr: XMLPlaylistParser, key: str) -> int:
            if not (value := psr.xml_source.get(key)):
                return 0
            return sum(1 for p in value.split("|") if p)

        return SyncResultXAutoPF(
            start=initial_count,
            start_included=_get_paths_sum(initial, "ExceptionsInclude"),
            start_excluded=_get_paths_sum(initial, "Exceptions"),
            start_compared=len(initial.xml_source["Conditions"].get("Condition", [])),
            start_limiter=initial.xml_source["Limit"].get("@Enabled", "False") == "True",
            start_sorter=len(initial.xml_source.get("SortBy", initial.xml_source.get("DefinedSort", []))) > 0,
            final=len(self.tracks),
            final_included=_get_paths_sum(parser, "ExceptionsInclude"),
            final_excluded=_get_paths_sum(parser, "Exceptions"),
            final_compared=len(parser.xml_source["Conditions"].get("Condition", [])),
            final_limiter=parser.xml_source["Limit"].get("@Enabled", "False") == "True",
            final_sorter=len(parser.xml_source.get("SortBy", parser.xml_source.get("DefinedSort", []))) > 0,
        )


class XMLPlaylistParser(File, PrettyPrinter):

    __slots__ = ("_path", "path_mapper", "xml")

    # noinspection SpellCheckingInspection
    #: Map of MusicBee field key name to Field enum
    name_field_map = {
        "None": None,
        "Title": Fields.TITLE,
        "ArtistPeople": Fields.ARTIST,
        "Album": Fields.ALBUM,  # album ignoring articles like 'the' and 'a' etc.
        "Album Artist": Fields.ALBUM_ARTIST,
        "TrackNo": Fields.TRACK_NUMBER,
        "TrackCount": Fields.TRACK_TOTAL,
        "GenreSplits": Fields.GENRES,
        "Year": Fields.YEAR,  # could also be 'YearOnly'?
        "BeatsPerMin": Fields.BPM,
        "DiscNo": Fields.DISC_NUMBER,
        "DiscCount": Fields.DISC_TOTAL,
        # "": Fields.COMPILATION,  # unmapped for compare
        "Comment": Fields.COMMENTS,
        "FileDuration": Fields.LENGTH,
        "Rating": Fields.RATING,
        # "ComposerPeople": Fields.COMPOSER,  # currently not supported by this program
        # "Conductor": Fields.CONDUCTOR,  # currently not supported by this program
        # "Publisher": Fields.PUBLISHER,  # currently not supported by this program
        "FilePath": Fields.PATH,
        "FolderName": Fields.FOLDER,
        "FileName": Fields.FILENAME,
        "FileExtension": Fields.EXT,
        # "": Fields.SIZE,  # unmapped for compare
        "FileKind": Fields.TYPE,
        "FileBitrate": Fields.BIT_RATE,
        "BitDepth": Fields.BIT_DEPTH,
        "FileSampleRate": Fields.SAMPLE_RATE,
        "FileChannels": Fields.CHANNELS,
        # "": Fields.DATE_CREATED,  # unmapped for compare
        "FileDateModified": Fields.DATE_MODIFIED,
        "FileDateAdded": Fields.DATE_ADDED,
        "FileLastPlayed": Fields.LAST_PLAYED,
        "FilePlayCount": Fields.PLAY_COUNT,
    }
    #: Map of Field enum to MusicBee field key name
    field_name_map = {field: name for name, field in name_field_map.items()}

    #: Settings for custom sort codes.
    defined_sort: dict[int, Mapping[Field, bool]] = {
        6: {
            Fields.ALBUM: False,
            Fields.DISC_NUMBER: False,
            Fields.TRACK_NUMBER: False,
            Fields.FILENAME: False
        }
        # TODO: implement field_code 78 - manual order according to the order of tracks found
        #  in the MusicBee library file for a given playlist.
    }
    #: The default/manual sort code
    default_sort = 78
    #: The default setting to use for the GroupBy setting
    default_group_by = "track"

    @property
    def path(self) -> Path:
        return self._path

    @property
    def xml_smart_playlist(self) -> dict[str, Any]:
        """The smart playlist data part of the loaded XML playlist data"""
        return self.xml["SmartPlaylist"]

    @property
    def xml_source(self) -> dict[str, Any]:
        """The source data part of the loaded XML playlist data"""
        return self.xml_smart_playlist["Source"]

    @property
    def description(self):
        """The description value of the XML playlist data"""
        return self.xml_source["Description"]

    @description.setter
    def description(self, value: str | None):
        """The description value of the XML playlist data"""
        if value:
            self.xml_source["Description"] = value
        else:
            self.xml_source.pop("Description", None)

    def __init__(self, path: str | Path, path_mapper: PathMapper = PathMapper()):
        if xmltodict is None:
            raise MusifyImportError(f"Cannot create {self.__class__.__name__} object. Required modules: xmltodict")

        self._path = Path(path)
        #: Maps paths stored in the playlist file.
        self.path_mapper = path_mapper
        #: A map representation of the loaded XML playlist data
        self.xml: dict[str, Any] = {}

    async def load(self) -> None:
        """Load ``xml`` object from the disk"""
        with open(self.path, "r", encoding="utf-8") as file:
            self.xml: dict[str, Any] = xmltodict.parse(file.read())

    async def save(self, dry_run: bool = True, *_, **__) -> None:
        """Save ``xml`` object to the disk"""
        if dry_run:
            return
        with open(self.path, 'w', encoding="utf-8") as file:
            xml_str = xmltodict.unparse(self.xml, pretty=True, short_empty_elements=True)
            file.write(xml_str.replace("/>", " />").replace('\t', '  '))

    def _get_comparer(self, xml: Mapping[str, Any]) -> Comparer:
        """
        Initialise and return a :py:class:`Comparer` from the relevant chunk of settings in ``xml`` playlist data.

        :param xml: The relevant chunk to generate a single :py:class:`Comparer` as found in
            the loaded XML object for this playlist.
            This function expects to be given only the XML part related to one Comparer condition.
        :return: The initialised :py:class:`Comparer`.
        """
        field_name = xml.get("@Field", "None")
        field: Field = self.name_field_map.get(field_name)
        if field is None:
            raise FieldError("Unrecognised field name", field=field_name)

        expected: tuple[str, ...] | None = tuple(v for k, v in xml.items() if k.startswith("@Value"))
        reference_required = next(iter(expected), None) == "[playing track]"
        if len(expected) == 0 or expected[0] == "[playing track]":
            expected = None

        return Comparer(
            condition=xml["@Comparison"], expected=expected, field=field, reference_required=reference_required
        )

    def _get_xml_from_comparer(self, comparer: Comparer | None = None) -> dict[str, Any]:
        """Parse the given ``comparer`` to its XML playlist representation."""
        if comparer is None:  # default value
            return {"@Field": "ArtistPeople", "@Comparison": "StartsWith", "@Value": ""}

        field_name = self.field_name_map.get(comparer.field)
        if field_name is None:
            raise FieldError("Unrecognised field", field=comparer.field)

        condition = "[playing track]" if comparer.reference_required else self._snake_to_pascal(comparer.condition)

        xml: dict[str, Any] = {"@Field": field_name, "@Comparison": condition}

        if comparer.expected is None:
            pass
        elif len(comparer.expected) == 0:
            xml["@Value"] = ""
        elif len(comparer.expected) == 1:
            xml["@Value"] = str(comparer.expected[0])
        else:
            for i, value in enumerate(comparer.expected, 1):
                xml[f"@Value{i}"] = str(value)

        return xml

    def get_matcher(self) -> AutoMatcher:
        """Initialise and return a :py:class:`FilterMatcher` object from loaded XML playlist data."""
        def get_exceptions(key: str) -> set[Path]:
            """Get exception paths from XML to include/exclude even if they meet/don't meet match compare conditions"""
            if not (raw_str := self.xml_source.get(key)):
                return set()

            paths = set(raw_str.split("|"))
            return set(map(Path, self.path_mapper.map_many(paths, check_existence=True)))

        include = get_exceptions("ExceptionsInclude")
        exclude = get_exceptions("Exceptions")

        comparers: dict[Comparer, tuple[bool, FilterComparers]] = {}
        for condition in to_collection(self.xml_source["Conditions"]["Condition"]):
            if any(key in condition for key in {"And", "Or"}):
                combine = "And" in condition
                conditions = condition["And" if combine else "Or"]
                sub_filter = FilterComparers(
                    comparers=[self._get_comparer(sub) for sub in to_collection(conditions["Condition"])],
                    match_all=conditions["@CombineMethod"] == "All"
                )
            else:
                combine = False
                sub_filter = FilterComparers()

            comparers[self._get_comparer(xml=condition)] = (combine, sub_filter)

        dummy_conditions = {"starts_with", "contains"}
        if len(comparers) == 1 and not next(iter(comparers.values()))[1].ready:
            # when user has not set an explicit comparer, a single empty 'allow all' comparer is assigned
            # check for this 'allow all' comparer and remove it if present to speed up comparisons
            c = next(iter(comparers))
            is_dummy_condition = any(cond in c.condition.casefold() for cond in dummy_conditions)
            if is_dummy_condition and len(c.expected) == 1 and not c.expected[0]:
                comparers = {}

        filter_include = FilterDefinedList[LocalTrack](values=include)
        filter_exclude = FilterDefinedList[LocalTrack](values=exclude)
        filter_compare = FilterComparers[LocalTrack](
            comparers, match_all=self.xml_source["Conditions"]["@CombineMethod"] == "All"
        )

        filter_include.transform = lambda x: Path(self.path_mapper.map(x, check_existence=False))
        filter_exclude.transform = lambda x: Path(self.path_mapper.map(x, check_existence=False))

        group_by_value = self._pascal_to_snake(self.xml_smart_playlist["@GroupBy"])
        group_by = None if group_by_value == self.default_group_by else TagFields.from_name(group_by_value)[0]

        return FilterMatcher(
            include=filter_include, exclude=filter_exclude, comparers=filter_compare, group_by=group_by
        )

    def parse_matcher(self, matcher: FilterMatcher | None = None) -> None:
        """
        Update the loaded ``xml`` object by parsing the given ``matcher`` to its XML playlist representation.
        Does not extract exception paths (i.e. include/exclude attributes).
        """
        if matcher is None or not matcher.ready:
            self.xml_smart_playlist["@GroupBy"] = self.default_group_by
            self.xml_source["Conditions"] = {"@CombineMethod": "All"} | {"Condition": self._get_xml_from_comparer()}
            return

        group_by = matcher.group_by.name.lower() if matcher.group_by else self.default_group_by
        self.xml_smart_playlist["@GroupBy"] = group_by.lower()
        self.xml_source["Conditions"] = self._parse_filter_comparer(matcher.comparers)

    def _parse_filter_comparer(self, filter_: FilterComparers) -> dict[str, Any]:
        combine_method = "All" if filter_.match_all else "Any"
        conditions: list[dict[str, Any]] = []
        for comparer, (combine, sub_filter) in filter_.comparers.items():
            condition = self._get_xml_from_comparer(comparer)
            if sub_filter.ready:
                combine_key = "And" if combine else "Or"
                condition[combine_key] = self._parse_filter_comparer(sub_filter)

            conditions.append(condition)

        if len(conditions) == 0:  # assign the default value when no comparers present
            conditions.append(self._get_xml_from_comparer())

        return {
            "@CombineMethod": combine_method,
            "Condition": conditions[0] if len(conditions) == 1 else conditions
        }

    def parse_exception_paths(
            self, matcher: FilterMatcher | None, items: list[File], original: list[File | MusifyItem]
    ) -> None:
        """
        Parse the exception paths (i.e. include/exclude attributes) for the given ``matcher``
        to its XML playlist representation. Does not extract any other attributes from the ``matcher``.

        :param matcher: The :py:class:`FilterMatcher` to parse.
        :param items: The items to export.
        :param original: The original items matched from the settings in the original file.
        """
        if matcher is None:
            self.xml_source.pop("ExceptionsInclude", None)
            self.xml_source.pop("Exceptions", None)
            return

        if not isinstance(matcher.include, FilterDefinedList) and not isinstance(matcher.exclude, FilterDefinedList):
            matcher.logger.warning(
                "Cannot export this filter to XML: Include and Exclude settings must both be list filters"
            )
            return

        items_mapped: dict[Path, File] = {item.path: item for item in items}

        if matcher.comparers:
            # match again on current conditions to check for differences from original list
            # which ensures that the paths included in the XML output
            # do not include paths that match any of the comparer or group_by conditions

            # copy the list of tracks as the sorter will modify the list order
            original = original.copy()
            # get the last played track as reference in case comparer is looking for the playing tracks as reference
            ItemSorter.sort_by_field(original, field=Fields.LAST_PLAYED, reverse=True)

            matched_mapped: dict[Path, File] = {
                item.path: item for item in matcher.comparers(original, reference=next(iter(original), None))
            } if matcher.comparers.ready else {}
            # noinspection PyProtectedMember
            matched_mapped |= {
                item.path: item for item in matcher._get_group_by_results(original, matched_mapped.values())
            }

            # get new include/exclude paths based on the leftovers after matching on comparers and group_by settings
            matcher.exclude.values = list(matched_mapped.keys() - items_mapped)
            matcher.include.values = [v for v in list(items_mapped - matched_mapped.keys()) if v not in matcher.exclude]
        else:
            matched_mapped = items_mapped

        include_items = tuple(items_mapped[path] for path in matcher.include if path in items_mapped)
        exclude_items = tuple(matched_mapped[path] for path in matcher.exclude if path in matched_mapped)

        if len(include_items) > 0:
            include_paths = self.path_mapper.unmap_many(include_items, check_existence=False)
            self.xml_source["ExceptionsInclude"] = "|".join(include_paths)
        if len(exclude_items) > 0:
            exclude_paths = self.path_mapper.unmap_many(exclude_items, check_existence=False)
            self.xml_source["Exceptions"] = "|".join(exclude_paths)

    def get_limiter(self) -> ItemLimiter | None:
        """Initialise and return a :py:class:`ItemLimiter` object from loaded XML playlist data."""
        conditions: Mapping[str, str] = self.xml_source["Limit"]
        if conditions["@Enabled"] != "True":
            return

        # MusicBee appears to have some extra allowance on time and byte limits of ~1.25
        return ItemLimiter(
            limit=int(conditions["@Count"]),
            on=LimitType.from_name(conditions["@Type"])[0],
            sorted_by=conditions["@SelectedBy"],
            allowance=1.25
        )

    @property
    def limiter_deduplication(self) -> bool:
        """This setting controls whether duplicates should be filtered out before running limiter operations."""
        return self.xml_source["Limit"]["@FilterDuplicates"] == "True"

    def parse_limiter(self, limiter: ItemLimiter | None = None, deduplicate: bool = False) -> None:
        """Update the loaded ``xml`` object by parsing the given ``limiter`` to its XML playlist representation."""
        if limiter is None:  # default value
            xml = {
                "@FilterDuplicates": str(deduplicate).title(),
                "@Enabled": "False",
                "@Count": "25",
                "@Type": "Items",
                "@SelectedBy": "Random"
            }
        else:
            xml = {
                "@FilterDuplicates": str(deduplicate).title(),
                "@Enabled": "True",
                "@Count": str(limiter.limit_max),
                "@Type": limiter.kind.name.title(),
                "@SelectedBy": self._snake_to_pascal(limiter.limit_sort)
            }

        self.xml_source["Limit"] = xml

    def get_sorter(self) -> ItemSorter | None:
        """Initialise and return a :py:class:`ItemLimiter` object from loaded XML playlist data."""
        fields: Sequence[Field] | Mapping[Field | bool] = ()

        if "SortBy" in self.xml_source:
            field_code = int(self.xml_source["SortBy"].get("@Field", 0))
        elif "DefinedSort" in self.xml_source:
            field_code = int(self.xml_source["DefinedSort"]["@Id"])
        else:
            return

        if field_code in self.defined_sort:
            fields = self.defined_sort[field_code]
            return ItemSorter(fields=fields)
        elif field_code != self.default_sort:
            field = Fields.from_value(field_code)[0]

            if "SortBy" in self.xml_source:
                fields = {field: self.xml_source["SortBy"]["@Order"] == "Descending"}
            elif "DefinedSort" in self.xml_source:
                fields = [field]
            else:
                raise SorterProcessorError("Sort type in XML not recognised")

        shuffle_mode_value = self._pascal_to_snake(self.xml_smart_playlist["@ShuffleMode"])
        if not fields and shuffle_mode_value != "none":
            shuffle_mode = ShuffleMode.from_name(shuffle_mode_value)[0]
            shuffle_weight = float(self.xml_smart_playlist.get("@ShuffleSameArtistWeight", 0))

            return ItemSorter(fields=fields, shuffle_mode=shuffle_mode, shuffle_weight=shuffle_weight)

        # TODO: remove defined_sort workaround here, should use manual sort when no `fields` - see self.defined_sort
        return ItemSorter(fields=fields or next(iter(self.defined_sort.values())))

    def parse_sorter(self, sorter: ItemSorter | None = None) -> None:
        """Update the loaded ``xml`` object by parsing the given ``sorter`` to its XML playlist representation."""
        self.xml_source.pop("SortBy", None)
        self.xml_source.pop("DefinedSort", None)

        if sorter is None:  # default value
            self.xml_smart_playlist["@ShuffleMode"] = "None"
            self.xml_smart_playlist["@ShuffleSameArtistWeight"] = "0.5"
            self.xml_source["SortBy"] = {"@Field": str(self.default_sort), "@Order": "Ascending"}
            return

        shuffle_mode = "None" if sorter.shuffle_mode is None else self._snake_to_pascal(sorter.shuffle_mode.name)
        self.xml_smart_playlist["@ShuffleMode"] = shuffle_mode
        self.xml_smart_playlist["@ShuffleSameArtistWeight"] = str(sorter.shuffle_weight)

        defined_sort_key = next((key for key, value in self.defined_sort.items() if value == sorter.sort_fields), None)
        if defined_sort_key is not None:
            self.xml_source["DefinedSort"] = {"@Id": str(defined_sort_key)}
            return

        if len(sorter.sort_fields) > 1:
            raise SorterProcessorError(
                "Cannot generate an XML representation of a mapping of many sort fields unless they have "
                "a defined sort ID. To parse these fields to XML, define this map in the 'defined_sort' "
                f"class attribute of this parser | {sorter.sort_fields}"
            )

        field, reverse_sort = next(iter(sorter.sort_fields.items()), (None, False))
        self.xml_source["SortBy"] = {
            "@Field": str(field.value if field is not None else self.default_sort),
            "@Order": "Descending" if reverse_sort else "Ascending"
        }

    def as_dict(self):
        return {"path": self.path, "path_mapper": self.path_mapper}



"""
Meta-operations for all playlists/tracks etc. stored in a given library.

A library can be a simple set of folders containing playlists/tracks etc.,
or a defined settings folder for any of the supported library managers.

Specific library types should implement :py:class:`LocalLibrary`.
"""
from .library import LocalLibrary
from .musicbee import MusicBee

LIBRARY_CLASSES = frozenset({LocalLibrary, MusicBee})



"""
Convert and validate Spotify ID and item types.
"""
from collections.abc import Mapping
from typing import Any

from aiorequestful.types import URLInput
from yarl import URL

from musify.exception import MusifyEnumError
from musify.libraries.core.collection import MusifyCollection
from musify.libraries.remote.core import RemoteResponse
from musify.libraries.remote.core.exception import RemoteError, RemoteIDTypeError, RemoteObjectTypeError
from musify.libraries.remote.core.types import APIInputValueSingle, APIInputValueMulti, RemoteIDType, RemoteObjectType
from musify.libraries.remote.core.wrangle import RemoteDataWrangler
from musify.libraries.remote.spotify import SOURCE_NAME
from musify.utils import to_collection


class SpotifyDataWrangler(RemoteDataWrangler):

    __slots__ = ()

    source = SOURCE_NAME
    unavailable_uri_dummy = "spotify:track:unavailable"
    url_api = URL("https://api.spotify.com/v1")
    url_ext = URL("https://open.spotify.com")

    @classmethod
    def get_id_type(cls, value: URLInput, kind: RemoteObjectType | None = None) -> RemoteIDType:
        value = str(value).strip().casefold()
        uri_split = value.split(':')

        if value.startswith(str(cls.url_api)):
            return RemoteIDType.URL
        elif value.startswith(str(cls.url_ext)):
            return RemoteIDType.URL_EXT
        elif len(uri_split) == RemoteIDType.URI.value and uri_split[0] == "spotify":  # URI
            if uri_split[1] == "user":
                return RemoteIDType.URI
            elif uri_split[1] != "user" and len(uri_split[2]) == RemoteIDType.ID.value:
                return RemoteIDType.URI
        elif len(value) == RemoteIDType.ID.value or kind == RemoteObjectType.USER:
            return RemoteIDType.ID
        raise RemoteIDTypeError(f"Could not determine ID type of given value: {value}")

    @classmethod
    def validate_id_type(cls, value: URLInput, kind: RemoteIDType = RemoteIDType.ALL) -> bool:
        value = str(value).strip().casefold()

        if kind == RemoteIDType.URL:
            return value.startswith(str(cls.url_api))
        elif kind == RemoteIDType.URL_EXT:
            return value.startswith(str(cls.url_ext))
        elif kind == RemoteIDType.URI:
            uri_split = value.split(':')
            if len(uri_split) != RemoteIDType.URI.value or uri_split[0] != "spotify":
                return False
            return uri_split[1] == "user" or (uri_split[1] != "user" and len(uri_split[2]) == RemoteIDType.ID.value)
        elif kind == RemoteIDType.ID:
            return len(value) == RemoteIDType.ID.value
        elif kind == RemoteIDType.ALL:
            try:
                cls.get_id_type(value)
                return True
            except RemoteIDTypeError:
                pass
        return False

    @classmethod
    def _get_item_type(
            cls, value: APIInputValueSingle[RemoteResponse], kind: RemoteObjectType | None = None
    ) -> RemoteObjectType | None:
        if isinstance(value, RemoteResponse):
            return cls._get_item_type_from_response(value)
        if isinstance(value, Mapping):
            return cls._get_item_type_from_mapping(value)

        value = str(value).strip()
        uri_check = value.split(':')

        if value.startswith(str(cls.url_api)) or value.startswith(str(cls.url_ext)):  # open/API URL
            value = value.removeprefix(str(cls.url_api if value.startswith(str(cls.url_api)) else cls.url_ext))
            url_path = URL(value).path.split("/")
            for chunk in url_path:
                try:
                    return RemoteObjectType.from_name(chunk.casefold().rstrip('s'))[0]
                except MusifyEnumError:
                    continue
        elif len(uri_check) == RemoteIDType.URI.value and uri_check[0].casefold() == "spotify":
            return RemoteObjectType.from_name(uri_check[1])[0]
        elif len(value) == RemoteIDType.ID.value or kind == RemoteObjectType.USER:
            # in these cases, we have to go on faith...
            return kind
        raise RemoteObjectTypeError(f"Could not determine item type of given value: {value}")

    @classmethod
    def _get_item_type_from_response(cls, value: RemoteResponse) -> RemoteObjectType:
        response_kind = cls._get_item_type_from_mapping(value.response)
        if value.kind != response_kind:
            raise RemoteObjectTypeError(
                f"RemoteResponse kind != actual response kind: {value.kind} != {response_kind}"
            )
        return value.kind

    @classmethod
    def _get_item_type_from_mapping(cls, value: Mapping[str, Any]) -> RemoteObjectType:
        if value.get("is_local", False):
            raise RemoteObjectTypeError("Cannot process local items")
        if "type" not in value:
            raise RemoteObjectTypeError(f"Given map does not contain a 'type' key: {value}")
        return RemoteObjectType.from_name(value["type"].casefold().rstrip('s'))[0]

    @classmethod
    def convert(
            cls,
            value: URLInput,
            kind: RemoteObjectType | None = None,
            type_in: RemoteIDType = RemoteIDType.ALL,
            type_out: RemoteIDType = RemoteIDType.ID
    ) -> str:
        if cls.validate_id_type(value, kind=type_out):
            return value
        if type_in == RemoteIDType.ALL or not cls.validate_id_type(value, kind=type_in):
            type_in = cls.get_id_type(value, kind=kind)

        value = str(value).strip()
        kind, id_ = cls._get_id(value=value, kind=kind, type_in=type_in)

        # reformat
        if type_out != RemoteIDType.ID:
            item = kind.name.lower().rstrip('s')
            if type_out == RemoteIDType.URL:
                return f"{cls.url_api}/{item}s/{id_}"
            elif type_out == RemoteIDType.URL_EXT:
                return f"{cls.url_ext}/{item}/{id_}"
            elif type_out == RemoteIDType.URI:
                return f"spotify:{item}:{id_}"
            raise RemoteIDTypeError

        return id_

    @classmethod
    def _get_id(
            cls, value: URLInput, kind: RemoteObjectType | None = None, type_in: RemoteIDType = RemoteIDType.ALL
    ) -> tuple[RemoteObjectType, str]:
        if isinstance(value, URL) or type_in == RemoteIDType.URL_EXT or type_in == RemoteIDType.URL:
            try:
                kind, id_ = cls._get_id_from_url(value=value, kind=kind)
            except StopIteration:
                raise RemoteObjectTypeError(f"Could not get ID from given URL: {value}")
        elif type_in == RemoteIDType.URI:
            try:
                kind, id_ = cls._get_id_from_uri(value=value)
            except StopIteration:
                raise RemoteObjectTypeError(f"Could not get ID from given URI: {value}")
        elif type_in == RemoteIDType.ID:
            if kind is None:
                raise RemoteIDTypeError("Input value is an ID and no defined 'kind' has been given.", RemoteIDType.ID)
            id_ = value
        else:
            raise RemoteIDTypeError(f"Could not determine item type: {value}")

        return kind, id_

    @classmethod
    def _get_id_from_url(cls, value: URLInput, kind: RemoteObjectType | None = None) -> tuple[RemoteObjectType, str]:
        url_path = URL(value).path.split("/")
        for chunk in url_path:
            try:
                kind = RemoteObjectType.from_name(chunk.rstrip('s'))[0]
                break
            except MusifyEnumError:
                continue

        if kind == RemoteObjectType.USER:
            name = kind.name.lower()
            try:
                id_ = url_path[url_path.index(name) + 1]
            except ValueError:
                id_ = url_path[url_path.index(name + "s") + 1]
        else:
            id_ = next(p for p in url_path if len(p) == RemoteIDType.ID.value)

        return kind, id_

    @classmethod
    def _get_id_from_uri(cls, value: str) -> tuple[RemoteObjectType, str]:
        uri_split = value.split(':')
        kind = RemoteObjectType.from_name(uri_split[1])[0]
        id_ = uri_split[2]
        return kind, id_

    @classmethod
    def extract_ids(cls, values: APIInputValueMulti[RemoteResponse], kind: RemoteObjectType | None = None) -> list[str]:
        def extract_id(value: APIInputValueSingle[RemoteResponse]) -> str:
            """Extract an ID from a given ``value``"""
            if isinstance(value, URLInput):
                return cls.convert(value, kind=kind, type_out=RemoteIDType.ID)
            elif isinstance(value, Mapping) and "id" in value:
                return value["id"]
            elif isinstance(value, RemoteResponse):
                return value.id

            raise RemoteError(f"Could not extract ID. Input data not recognised: {type(value)}")

        values = to_collection(values) if not isinstance(values, MusifyCollection) else [values]
        return [extract_id(value=value) for value in to_collection(values)]



"""
Mixin for all implementations of :py:class:`RemoteAPI` for the Spotify API.

Also includes the default arguments to be used when requesting authorisation from the Spotify API.
"""
from http import HTTPMethod
from pathlib import Path

from aiohttp import ClientResponse
from aiorequestful.auth.oauth2 import AuthorisationCodeFlow
from aiorequestful.auth.utils import AuthRequest
from aiorequestful.cache.backend.base import ResponseCache, ResponseRepository
from aiorequestful.cache.session import CachedSession
from aiorequestful.types import URLInput
from yarl import URL

from musify.libraries.remote.core.exception import APIError
from musify.libraries.remote.spotify.api.cache import SpotifyRepositorySettings, SpotifyPaginatedRepositorySettings
from musify.libraries.remote.spotify.api.item import SpotifyAPIItems
from musify.libraries.remote.spotify.api.misc import SpotifyAPIMisc
from musify.libraries.remote.spotify.api.playlist import SpotifyAPIPlaylists
from musify.libraries.remote.spotify.wrangle import SpotifyDataWrangler
from musify.types import UnitIterable


class SpotifyAPI(SpotifyAPIMisc, SpotifyAPIItems, SpotifyAPIPlaylists):
    """
    Collection of endpoints for the Spotify API.

    :param client_id: The client ID to use when authorising requests.
    :param client_secret: The client secret to use when authorising requests.
    :param scope: The scopes to request access to.
    :param cache: When given, attempt to use this cache for certain request types before calling the API.
    :param token_file_path: Optionally, provide a path to save/load a response token.
    """

    __slots__ = ()

    @property
    def user_id(self) -> str | None:
        """ID of the currently authenticated user"""
        if not self.user_data:
            raise APIError(
                "User data not set. Either set explicitly or enter the "
                f"{self.__class__.__name__} context to set automatically."
            )
        return self.user_data["id"]

    @property
    def user_name(self) -> str | None:
        """Name of the currently authenticated user"""
        if not self.user_data:
            raise APIError(
                "User data not set. Either set explicitly or enter the "
                f"{self.__class__.__name__} context to set automatically."
            )
        return self.user_data["display_name"]

    _url_auth = URL.build(scheme="https", host="accounts.spotify.com")

    def __init__(
            self,
            client_id: str | None = None,
            client_secret: str | None = None,
            scope: UnitIterable[str] = (),
            cache: ResponseCache | None = None,
            token_file_path: str | Path = None,
    ):
        wrangler = SpotifyDataWrangler()
        authoriser = AuthorisationCodeFlow.create_with_encoded_credentials(
            service_name=wrangler.source,
            user_request_url=self._url_auth.with_path("authorize"),
            token_request_url=self._url_auth.with_path("api/token"),
            refresh_request_url=self._url_auth.with_path("api/token"),
            client_id=client_id,
            client_secret=client_secret,
            scope=scope,
        )

        if not hasattr(authoriser.token_request, "headers"):
            authoriser.token_request.headers = {}
        authoriser.token_request.headers["content-type"] = "application/x-www-form-urlencoded"

        if not hasattr(authoriser.refresh_request, "headers"):
            authoriser.refresh_request.headers = {}
        authoriser.refresh_request.headers["content-type"] = "application/x-www-form-urlencoded"

        if token_file_path:
            authoriser.response.file_path = Path(token_file_path)
        authoriser.response.additional_headers = {
            "Accept": "application/json", "Content-Type": "application/json"
        }

        authoriser.tester.request = AuthRequest(
            method=HTTPMethod.GET, url=wrangler.url_api.joinpath("me")
        )
        authoriser.tester.response_test = self._response_test
        authoriser.tester.max_expiry = 600

        super().__init__(authoriser=authoriser, wrangler=wrangler, cache=cache)

    async def _response_test(self, response: ClientResponse) -> bool:
        r = await response.json()
        return self.url_key in r and "display_name" in r

    # noinspection PyAsyncCall
    async def _setup_cache(self) -> None:
        session = self.handler.session
        if not isinstance(session, CachedSession):
            return

        cache = session.cache
        cache.repository_getter = self._get_cache_repository

        cache.create_repository(SpotifyRepositorySettings(name="tracks"))
        cache.create_repository(SpotifyRepositorySettings(name="audio_features"))
        cache.create_repository(SpotifyRepositorySettings(name="audio_analysis"))

        cache.create_repository(SpotifyRepositorySettings(name="albums"))
        cache.create_repository(SpotifyPaginatedRepositorySettings(name="album_tracks"))

        cache.create_repository(SpotifyRepositorySettings(name="artists"))
        cache.create_repository(SpotifyPaginatedRepositorySettings(name="artist_albums"))

        cache.create_repository(SpotifyRepositorySettings(name="shows"))
        cache.create_repository(SpotifyRepositorySettings(name="episodes"))
        cache.create_repository(SpotifyPaginatedRepositorySettings(name="show_episodes"))

        cache.create_repository(SpotifyRepositorySettings(name="audiobooks"))
        cache.create_repository(SpotifyRepositorySettings(name="chapters"))
        cache.create_repository(SpotifyPaginatedRepositorySettings(name="audiobook_chapters"))

        await cache

    @staticmethod
    def _get_cache_repository(cache: ResponseCache, url: URLInput) -> ResponseRepository | None:
        path = URL(url).path
        path_split = [part.replace("-", "_") for part in path.split("/")[2:]]

        if len(path_split) < 3:
            name = path_split[0]
        else:
            name = "_".join([path_split[0].rstrip("s"), path_split[2].rstrip("s") + "s"])

        return cache.get(name)



"""
The fundamental core collection classes for the entire package.
"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from collections.abc import MutableSequence, Iterable, Mapping, Collection
from dataclasses import dataclass
from pathlib import Path
from typing import Any, SupportsIndex, Self

from yarl import URL

from musify.base import MusifyObject, MusifyItem, HasLength
from musify.exception import MusifyTypeError, MusifyKeyError, MusifyAttributeError
from musify.field import Field
from musify.file.base import File
from musify.libraries.remote.core import RemoteResponse
from musify.processors.sort import ShuffleMode, ItemSorter
from musify.types import UnitSequence

type ItemGetterTypes = str | URL | MusifyItem | Path | File | RemoteResponse


@dataclass
class ItemGetterStrategy[KT](metaclass=ABCMeta):
    """Abstract base class for strategies relating to __getitem__ operations on a :py:class:`MusifyCollection`"""

    key: KT

    @property
    @abstractmethod
    def name(self) -> str:
        """The name to assign to this ItemGetter when logging"""
        raise NotImplementedError

    @abstractmethod
    def get_value_from_item(self, item: MusifyItem) -> KT:
        """Retrieve the appropriate value from a given ``item`` for this ItemGetter type"""
        raise NotImplementedError

    def get_item[IT](self, collection: MusifyCollection[IT]) -> IT:
        """Run this strategy and return the matched item from the given ``collection``"""
        try:
            return next(item for item in collection.items if self.get_value_from_item(item) == self.key)
        except AttributeError:
            raise MusifyAttributeError(f"Items in collection do not have the attribute {self.name!r}")
        except StopIteration:
            raise MusifyKeyError(f"No matching item found for {self.name}: {self.key}")


class NameGetter(ItemGetterStrategy):
    """Get an item via its name for a :py:class:`MusifyCollection`"""
    @property
    def name(self) -> str:
        return "name"

    def get_value_from_item(self, item: MusifyObject) -> str:
        return item.name


class PathGetter(ItemGetterStrategy):
    """Get an item via its path for a :py:class:`MusifyCollection`"""
    @property
    def name(self) -> str:
        return "path"

    def get_value_from_item(self, item: File) -> Path:
        return item.path


class RemoteIDGetter(ItemGetterStrategy):
    """Get an item via its remote ID for a :py:class:`MusifyCollection`"""
    @property
    def name(self) -> str:
        return "remote ID"

    def get_value_from_item(self, item: RemoteResponse) -> str:
        return item.id


class RemoteURIGetter(ItemGetterStrategy):
    """Get an item via its remote URI for a :py:class:`MusifyCollection`"""
    @property
    def name(self) -> str:
        return "URI"

    def get_value_from_item(self, item: MusifyItem | RemoteResponse) -> str:
        return item.uri


class RemoteURLAPIGetter(ItemGetterStrategy):
    """Get an item via its remote API URL for a :py:class:`MusifyCollection`"""
    @property
    def name(self) -> str:
        return "API URL"

    def get_value_from_item(self, item: RemoteResponse) -> URL:
        return item.url


class RemoteURLEXTGetter(ItemGetterStrategy):
    """Get an item via its remote external URL for a :py:class:`MusifyCollection`"""
    @property
    def name(self) -> str:
        return "external URL"

    def get_value_from_item(self, item: RemoteResponse) -> URL:
        return item.url_ext


class MusifyCollection[T: MusifyItem](MusifyObject, MutableSequence[T], HasLength):
    """Generic class for storing a collection of musify items."""

    __slots__ = ()

    @property
    @abstractmethod
    def items(self) -> list[T]:
        """The items in this collection"""
        raise NotImplementedError

    @property
    @abstractmethod
    def length(self) -> float | None:
        """Total duration of all items in this collection in seconds"""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def _validate_item_type(items: Any | Iterable[Any]) -> bool:
        """
        Validate the given :py:class:`MusifyItem` by ensuring it matches the allowed item type for this collection.
        Used to validate input :py:class:`MusifyItem` types given to functions that
        modify the stored items in this collection.

        :param items: The item or items to validate
        :return: True if valid, False if not.
        """
        raise NotImplementedError

    def count(self, __item: T) -> int:
        """Return the number of occurrences of the given :py:class:`MusifyItem` in this collection"""
        if not self._validate_item_type(__item):
            raise MusifyTypeError(type(__item).__name__)
        return self.items.count(__item)

    def index(self, __item: T, __start: SupportsIndex = None, __stop: SupportsIndex = None) -> int:
        """
        Return first index of item from items in this collection.

        :raise ValueError: If the value is not present.
        :raise MusifyTypeError: If given item does not match the item type of this collection.
        """
        if not self._validate_item_type(__item):
            raise MusifyTypeError(type(__item).__name__)
        return self.items.index(__item, __start or 0, __stop or len(self.items))

    def copy(self) -> list[T]:
        """Return a shallow copy of the list of items in this collection"""
        return self.items.copy()

    def append(self, __item: T, allow_duplicates: bool = True) -> None:
        """Append one item to the items in this collection"""
        if not self._validate_item_type(__item):
            raise MusifyTypeError(type(__item).__name__)
        if allow_duplicates or __item not in self.items:
            self.items.append(__item)

    def extend(self, __items: Iterable[T], allow_duplicates: bool = True) -> None:
        """Append many items to the items in this collection"""
        if not self._validate_item_type(__items):
            raise MusifyTypeError([type(i).__name__ for i in __items])
        if isinstance(__items, MusifyCollection):
            __items = __items.items

        if allow_duplicates:
            self.items.extend(__items)
        else:
            self.items.extend(item for item in __items if item not in self.items)

    def insert(self, __index: int, __item: T, allow_duplicates: bool = True) -> None:
        """Insert given :py:class:`MusifyItem` before the given index"""
        if not self._validate_item_type(__item):
            raise MusifyTypeError(type(__item))
        if allow_duplicates or __item not in self.items:
            self.items.insert(__index, __item)

    def remove(self, __item: T) -> None:
        """Remove one item from the items in this collection"""
        if not self._validate_item_type(__item):
            raise MusifyTypeError(type(__item))
        self.items.remove(__item)

    def pop(self, __item: SupportsIndex = None) -> T:
        """Remove one item from the items in this collection and return it"""
        return self.items.pop(__item) if __item else self.items.pop()

    def reverse(self) -> None:
        """Reverse the order of items in this collection in-place"""
        self.items.reverse()

    def clear(self) -> None:
        """Remove all items from this collection"""
        self.items.clear()

    def sort(
            self,
            fields: UnitSequence[Field | None] | Mapping[Field | None, bool] = (),
            shuffle_mode: ShuffleMode | None = None,
            shuffle_weight: float = 1.0,
            key: Field | None = None,
            reverse: bool = False,
    ) -> None:
        """
        Sort items in this collection in-place based on given conditions.
        If key is given,

        :param fields:
            * When None and ShuffleMode is RANDOM, shuffle the tracks. Otherwise, do nothing.
            * List of tags/properties to sort by.
            * Map of `{<tag/property>: <reversed>}`. If reversed is true, sort the ``tag/property`` in reverse.
        :param shuffle_mode: The mode to use for shuffling.
        :param shuffle_weight: The weights (between 0 and 1) to apply to shuffling modes that can use it.
            This value will automatically be limited to within the accepted range 0 and 1.
        :param key: Tag or property to sort on. Can be given instead of ``fields`` for a simple sort.
            If set, all other fields apart from ``reverse`` are ignored.
            If None, ``fields``, ``shuffle_mode``, ``shuffle_by``, and ``shuffle_weight`` are used to apply sorting.
        :param reverse: If true, reverse the order of the sort at the end.
        """
        if key is not None:
            ItemSorter.sort_by_field(self.items, field=key)
        else:
            ItemSorter(fields=fields, shuffle_mode=shuffle_mode, shuffle_weight=shuffle_weight)(self.items)

        if reverse:
            self.items.reverse()

    def intersection(self, other: Iterable[T]) -> list[T]:
        """
        Return the intersection between the items in this collection and an ``other`` collection as a new list.

        (i.e. all items that are in both this collection and the ``other`` collection).
        """
        return [item for item in self if item in other]

    def difference(self, other: Iterable[T]) -> list[T]:
        """
        Return the difference between the items in this collection and an ``other`` collection as a new list.

        (i.e. all items that are in this collection but not the ``other`` collection).
        """
        return [item for item in self if item not in other]

    def outer_difference(self, other: Iterable[T]) -> list[T]:
        """
        Return the outer difference between the items in this collection and an ``other`` collection as a new list.

        (i.e. all items that are in the ``other`` collection but not in this collection).
        """
        return [item for item in other if item not in self]

    @staticmethod
    def _condense_attributes(attributes: dict[str, Any]) -> dict[str, Any]:
        """Condense the attributes of the given map for cleaner attribute displaying"""
        def condense(key: str, value: Any) -> tuple[str, Any]:
            """Decide whether this key-value pair should be condensed and condense them."""
            if isinstance(value, Collection) and not isinstance(value, str):
                if any(isinstance(v, MusifyObject) for v in value) or len(value) > 20 or len(value) == 0:
                    return f"{key.rstrip("s")}_count", len(value)
            return key, value

        return dict(condense(k, v) for k, v in attributes.items())

    def as_dict(self):
        return self._condense_attributes(self._get_attributes())

    def _json_attributes(self):
        return self._get_attributes()

    def __eq__(self, __collection: MusifyCollection | Iterable[T]):
        """Names equal and all items equal in order"""
        name = self.name == __collection.name if isinstance(__collection, MusifyCollection) else True
        length = len(self) == len(__collection)
        items = all(x == y for x, y in zip(self, __collection))
        return name and length and items

    def __ne__(self, __collection: MusifyCollection | Iterable[T]):
        return not self.__eq__(__collection)

    def __len__(self):
        return len(self.items)

    def __iter__(self):
        return iter(self.items)

    def __reversed__(self):
        return reversed(self.items)

    def __contains__(self, __item: T):
        return any(__item == i for i in self.items)

    def __add__(self, __items: list[T] | Self):
        if isinstance(__items, MusifyCollection):
            return self.items + __items.items
        return self.items + __items

    def __iadd__(self, __items: Iterable[T]):
        self.extend(__items)
        return self

    def __sub__(self, __items: Iterable[T]):
        items = self.copy()
        for item in __items:
            items.remove(item)
        return items

    def __isub__(self, __items: Iterable[T]):
        if not isinstance(__items, Iterable):
            raise MusifyTypeError("You must provide an iterable object to use this functionality.")

        for item in __items:
            self.remove(item)
        return self

    def __getitem__(self, __key: int | slice | ItemGetterTypes) -> T | list[T] | list[T, None, None]:
        """
        Returns the item in this collection by matching on a given index/Item/URI/ID/URL.
        If an :py:class:`MusifyItem` is given, the URI is extracted from this item
        and the matching Item from this collection is returned.
        If a :py:class:`RemoteResponse` is given, the ID is extracted from this object
        and the matching RemoteResponse from this collection is returned.
        """
        if isinstance(__key, int) or isinstance(__key, slice):  # simply index the list or items
            return self.items[__key]

        getters = self.__get_item_getters(__key)
        if not getters:
            raise MusifyKeyError(f"Unrecognised key type | {__key=} | type={type(__key).__name__}")

        caught_exceptions = []
        for getter in getters:
            try:
                return getter.get_item(self)
            except (MusifyAttributeError, MusifyKeyError) as ex:
                caught_exceptions.append(ex)

        raise MusifyKeyError(
            f"Key is invalid. The following errors were thrown: {", ".join(map(str, caught_exceptions))}"
        )

    @classmethod
    def __get_item_getters(cls, __key: ItemGetterTypes) -> list[ItemGetterStrategy]:
        getters = []

        if isinstance(__key, File):
            getters.append(PathGetter(__key.path))
        if isinstance(__key, Path):
            getters.append(PathGetter(__key))
        if isinstance(__key, RemoteResponse):
            getters.append(RemoteIDGetter(__key.id))
        if isinstance(__key, MusifyItem) and __key.has_uri:
            getters.append(RemoteURIGetter(__key.uri))
        if isinstance(__key, URL):
            getters.append(RemoteURLAPIGetter(__key))
            getters.append(RemoteURLEXTGetter(__key))
        if isinstance(__key, MusifyObject):
            getters.append(NameGetter(__key.name))
        if isinstance(__key, str):
            getters.extend(cls.__get_item_getters_str(__key))

        return getters

    @staticmethod
    def __get_item_getters_str(__key: str) -> list[ItemGetterStrategy]:
        getters = []

        try:
            url = URL(__key)
            getters.append(RemoteURLAPIGetter(url))
            getters.append(RemoteURLEXTGetter(url))
        except TypeError:
            pass

        try:
            path = Path(__key)
            getters.append(PathGetter(path))
        except TypeError:
            pass

        getters.extend([
            RemoteIDGetter(__key),
            RemoteURIGetter(__key),
            NameGetter(__key),
        ])

        return getters

    def __setitem__(self, __key: str | int | T, __value: T):
        """Replace the item at a given ``__key`` with the given ``__value``."""
        try:
            item = self[__key]
        except KeyError:
            raise MusifyKeyError(f"Given index is out of range: {__key}")

        if type(__value) is not type(item):  # only merge attributes if matching types
            raise MusifyTypeError("Trying to set on mismatched item types")

        self.items[self.index(item)] = __value

    def __delitem__(self, __key: str | int | T):
        self.remove(__key)



"""
Core functionality for all remote music streaming services.

Defines framework for interacting with music streaming service provided APIs
with some implementations for the handling of this data.

Also defines abstract classes to represent objects derived from the types of
responses that can be returned by the API.
"""
from ._response import RemoteResponse



"""
Operations relating to reading and writing track data and properties to various types of playlist files.

Specific audio file types should implement :py:class:`LocalPlaylist`.
"""
from .base import LocalPlaylist
from .m3u import M3U
from .utils import PLAYLIST_CLASSES, PLAYLIST_FILETYPES, load_playlist
from .xautopf import XAutoPF



"""
The MP3 implementation of a :py:class:`LocalTrack`.
"""
from collections.abc import Iterable
from io import BytesIO
from typing import Any

import mutagen
import mutagen.id3
import mutagen.mp3

from musify.field import TagMap
from musify.file.image import open_image, get_image_bytes
from musify.libraries.local.exception import TagError
# noinspection PyProtectedMember
from musify.libraries.local.track._tags import TagReader, TagWriter
from musify.libraries.local.track.field import LocalTrackField
from musify.libraries.local.track.track import LocalTrack

try:
    from PIL import Image
except ImportError:
    Image = None


class _MP3TagReader(TagReader[mutagen.mp3.MP3]):

    __slots__ = ()

    def read_tag(self, tag_ids: Iterable[str]) -> list[Any] | None:
        # MP3 tag ids come in parts separated by : i.e. 'COMM:ID3v1 Comment:eng'
        # need to search all actual MP3 tag ids to check if the first part equals any of the given base tag ids
        tag_ids = tuple(mp3_id for mp3_id in self.file.keys() for tag_id in tag_ids if mp3_id.split(":")[0] == tag_id)

        values = []
        for tag_id in tag_ids:
            value = self.file.get(tag_id)
            if value is None:
                continue

            # convert id3 object to python types, causes downstream if not
            if isinstance(value, mutagen.id3.TextFrame):
                values.extend(str(value).split('\x00'))
            elif isinstance(value, mutagen.id3.APIC):
                values.append(value)
            else:
                raise TagError(f"Unrecognised id3 type: {value} ({type(value)})")

        return values if len(values) > 0 else None

    def read_genres(self) -> list[str] | None:
        """Extract metadata from file for genre"""
        values = self.read_tag(self.tag_map.genres)
        if values is None:
            return
        return [genre for value in values for genre in value.split(";")]

    def read_images(self):
        if Image is None:
            return

        values = self.read_tag(self.tag_map.images)
        return [Image.open(BytesIO(value.data)) for value in values] if values is not None else None


class _MP3TagWriter(TagWriter[mutagen.mp3.MP3]):

    __slots__ = ()

    def _clear_tag(self, tag_name: str, dry_run: bool = True) -> bool:
        removed = False

        tag_ids = self.tag_map[tag_name]
        if tag_ids is None or len(tag_ids) is None:
            return removed

        for tag_id_prefix in tag_ids:
            for mp3_id in list(self.file.keys()).copy():
                if mp3_id.split(":")[0] == tag_id_prefix and self.file[mp3_id]:
                    if not dry_run:
                        del self.file[mp3_id]
                    removed = True

        return removed

    def _write_tag(self, tag_id: str | None, tag_value: Any, dry_run: bool = True) -> bool:
        if not dry_run:
            self.file[tag_id] = getattr(mutagen.id3, tag_id)(3, text=str(tag_value))
        return True

    def _write_genres(self, track: LocalTrack, dry_run: bool = True) -> bool:
        values = ";".join(track.genres or [])
        return self.write_tag(next(iter(self.tag_map.genres), None), values, dry_run)

    def _write_images(self, track: LocalTrack, dry_run: bool = True) -> bool:
        tag_id_prefix = next(iter(self.tag_map.images), None)

        updated = False
        for image_kind, image_link in track.image_links.items():
            image = open_image(image_link)

            if not dry_run and tag_id_prefix is not None:
                image_kind_attr = image_kind.upper().replace(" ", "_")
                image_type: mutagen.id3.PictureType = getattr(mutagen.id3.PictureType, image_kind_attr)
                tag_id = f"{tag_id_prefix}:{image_kind}"

                # noinspection PyUnresolvedReferences
                self.file[tag_id] = mutagen.id3.APIC(
                    encoding=mutagen.id3.Encoding.UTF8,
                    # desc=image_kind,
                    mime=Image.MIME[image.format],
                    type=image_type,
                    data=get_image_bytes(image)
                )

            image.close()
            track.has_image = tag_id_prefix is not None or track.has_image
            updated = tag_id_prefix is not None

        return updated

    def _write_comments(self, track: LocalTrack, dry_run: bool = True) -> bool:
        tag_id_prefix = next(iter(self.tag_map.comments), None)
        self.delete_tags(tags=LocalTrackField.COMMENTS, dry_run=dry_run)

        for comment in track.comments:
            # noinspection PyUnresolvedReferences
            comm = mutagen.id3.COMM(
                encoding=mutagen.id3.Encoding.UTF8, desc="ID3v1 Comment", lang="eng", text=[comment]
            )
            # noinspection PyUnresolvedReferences
            tag_id = f"{tag_id_prefix}:{comm.desc}:{comm.lang}"
            if not dry_run and tag_id is not None:
                self.file[tag_id] = comm

        return tag_id_prefix is not None

    def _write_uri(self, track: LocalTrack, dry_run: bool = True) -> bool:
        if not self.remote_wrangler:
            return False

        tag_value = self.remote_wrangler.unavailable_uri_dummy if not track.has_uri else track.uri

        # if applying URI as comment, clear comments and add manually with custom description
        if self.uri_tag == LocalTrackField.COMMENTS:
            tag_id_prefix = next(iter(self.tag_map.comments), None)
            self.delete_tags(tags=self.uri_tag, dry_run=dry_run)

            # noinspection PyUnresolvedReferences
            comm = mutagen.id3.COMM(encoding=mutagen.id3.Encoding.UTF8, lang="eng", desc="URI", text=[tag_value])
            # noinspection PyUnresolvedReferences
            tag_id = f"{tag_id_prefix}:{comm.desc}:{comm.lang}"
            if not dry_run and tag_id is not None:
                self.file[tag_id] = comm

            return tag_id is not None
        else:
            tag_id = next(iter(self.tag_map[self.uri_tag.name.lower()]), None)
            return self.write_tag(tag_id, tag_value, dry_run)


class MP3(LocalTrack[mutagen.mp3.MP3, _MP3TagReader, _MP3TagWriter]):

    __slots__ = ()

    valid_extensions = frozenset({".mp3"})

    # noinspection SpellCheckingInspection
    #: Map of human-friendly tag name to ID3 tag ids for a given file type
    tag_map = TagMap(
        title=["TIT2"],
        artist=["TPE1"],
        album=["TALB"],
        album_artist=["TPE2"],
        track_number=["TRCK"],
        track_total=["TRCK"],
        genres=["TCON"],
        date=["TDRC", "TDAT", "TDOR"],
        year=["TYER", "TORY"],
        bpm=["TBPM"],
        key=["TKEY"],
        disc_number=["TPOS"],
        disc_total=["TPOS"],
        compilation=["TCMP"],
        comments=["COMM"],
        images=["APIC"],
    )

    def _create_reader(self, file: mutagen.mp3.MP3):
        return _MP3TagReader(file, tag_map=self.tag_map, remote_wrangler=self._remote_wrangler)

    def _create_writer(self, file: mutagen.mp3.MP3):
        return _MP3TagWriter(file, tag_map=self.tag_map, remote_wrangler=self._remote_wrangler)



"""
Implements all functionality pertaining to reading metadata/tags/properties for a :py:class:`LocalTrack`.
"""
import re
from abc import ABCMeta, abstractmethod
from collections.abc import Iterable
from typing import Any

import mutagen

from musify.libraries.local.track._tags.base import TagProcessor
from musify.libraries.remote.core.types import RemoteIDType
from musify.utils import to_collection

try:
    from PIL import Image
    ImageType = list[Image.Image] | None
except ImportError:
    Image = None
    ImageType = None


class TagReader[T: mutagen.FileType](TagProcessor, metaclass=ABCMeta):
    """Functionality for reading tags/metadata/properties from a mutagen object."""

    __slots__ = ()

    def read_tag(self, tag_ids: Iterable[str]) -> list[Any] | None:
        """Extract all tag values from file for a given list of tag IDs"""
        values = []
        for tag_id in tag_ids:
            value = self.file.get(tag_id)
            if value is None or (isinstance(value, str) and len(value.strip()) == 0):
                # skip null or empty/blank strings
                continue

            if isinstance(value, (list, set, tuple)) and all(isinstance(val, str) for val in value):
                values.extend(v for val in value for v in val.split('\x00'))
            elif isinstance(value, (list, set, tuple)):
                values.extend(value)
            elif isinstance(value, str):
                values.extend(value.split('\x00'))
            else:
                values.append(value)

        return values if len(values) > 0 else None

    def read_title(self) -> str | None:
        """Extract track title tags from file"""
        values = self.read_tag(self.tag_map.title)
        return str(values[0]) if values is not None else None

    def read_artist(self) -> str | None:
        """Extract artist tags from file"""
        values = self.read_tag(self.tag_map.artist)
        return str(values[0]) if values is not None else None

    def read_album(self) -> str | None:
        """Extract album tags from file"""
        values = self.read_tag(self.tag_map.album)
        return str(values[0]) if values is not None else None

    def read_album_artist(self) -> str | None:
        """Extract album artist tags from file"""
        values = self.read_tag(self.tag_map.album_artist)
        return str(values[0]) if values is not None else None

    def read_track_number(self) -> int | None:
        """Extract track number tags from file"""
        values = self.read_tag(self.tag_map.track_number)
        if values is None:
            return

        value = str(values[0])
        return int(value.split(self.num_sep)[0]) if self.num_sep in value else int(value)

    def read_track_total(self) -> int | None:
        """Extract total track count tags from file"""
        values = self.read_tag(self.tag_map.track_total)
        if values is None:
            return

        value = str(values[0])
        return int(value.split(self.num_sep)[1]) if self.num_sep in value else int(value)

    def read_genres(self) -> list[str] | None:
        """Extract genre tags from file"""
        values = self.read_tag(self.tag_map.genres)
        return list(map(str, values)) if values is not None else None

    def read_date(self) -> tuple[int | None, int | None, int | None] | None:
        """Extract year tags from file"""
        values = self.read_tag(self.tag_map.date)

        if values is None:  # attempt to read each part individually
            year = self.read_tag(self.tag_map.year)
            year = int(re.match(r"(\d{4})", str(year[0])).group(1)) if year else None
            month = self.read_tag(self.tag_map.month)
            month = int(re.match(r"(\d{1,2})", str(month[0])).group(1)) if month else None
            day = self.read_tag(self.tag_map.day)
            day = int(re.match(r"(\d{1,2})", str(day[0])).group(1)) if day else None
            return year, month, day
        elif 0 < len(values) <= 3 and all(str(value).isdigit() for value in values):
            values = ["/".join(values)]  # just join to string and allow later regex matches to determine format

        # YYYY-MM-DD
        match = re.match(r"(\d{4})\D+(\d{1,2})\D+(\d{1,2})", str(values[0]))
        if match:
            return int(match.group(1)), int(match.group(2)), int(match.group(3))

        # DD-MM-YY
        match = re.match(r"(\d{1,2})\D+(\d{1,2})\D+(\d{4})", str(values[0]))
        if match:
            return int(match.group(3)), int(match.group(2)), int(match.group(1))

        # YYYY-MM
        match = re.match(r"(\d{4})\D+(\d{1,2})", str(values[0]))
        if match:
            return int(match.group(1)), int(match.group(2)), None

        # MM-YYYY
        match = re.match(r"(\d{4})\D+(\d{1,2})", str(values[0]))
        if match:
            return int(match.group(2)), int(match.group(1)), None

        # YYYY
        match = re.match(r"(\d{4})", str(values[0]))
        if match:
            return int(match.group(1)), None, None

    def read_bpm(self) -> float | None:
        """Extract BPM tags from file"""
        values = self.read_tag(self.tag_map.bpm)
        try:
            return float(values[0]) if values is not None else None
        except ValueError:
            return None

    def read_key(self) -> str | None:
        """Extract key tags from file"""
        values = self.read_tag(self.tag_map.key)
        return str(values[0]) if values is not None else None

    def read_disc_number(self) -> int | None:
        """Extract disc number tags from file"""
        values = self.read_tag(self.tag_map.disc_number)
        if values is None:
            return

        value = str(values[0])
        return int(value.split(self.num_sep)[0]) if self.num_sep in value else int(value)

    def read_disc_total(self) -> int | None:
        """Extract total disc count tags from file"""
        values = self.read_tag(self.tag_map.disc_total)
        if values is None:
            return

        value = str(values[0])
        return int(value.split(self.num_sep)[1]) if self.num_sep in value else int(value)

    def read_compilation(self) -> bool | None:
        """Extract compilation tags from file"""
        values = self.read_tag(self.tag_map.compilation)
        try:
            return bool(int(values[0])) if values is not None else None
        except ValueError:
            return None

    def read_comments(self) -> list[str] | None:
        """Extract comment tags from file"""
        values = self.read_tag(self.tag_map.comments)
        return set(map(str, values)) if values is not None else None

    def read_uri(self) -> str | None:
        """Extract data relating to remote URI value from file"""
        wrangler = self.remote_wrangler
        if not wrangler:
            return

        read_method = getattr(self, f"read_{self.uri_tag.name.lower()}")
        possible_values: tuple[str, ...] | None = to_collection(read_method())
        if not possible_values:
            return None

        # WORKAROUND: for dodgy MP3 tag comments; split on null and take first value
        possible_values = tuple(val for values in possible_values for val in values.split("\x00"))
        for uri in possible_values:
            if uri == wrangler.unavailable_uri_dummy or wrangler.validate_id_type(uri, kind=RemoteIDType.URI):
                return uri

        return None

    @abstractmethod
    def read_images(self) -> ImageType:
        """Extract image from file"""
        raise NotImplementedError

    def check_for_images(self) -> bool:
        """Check if file has embedded images"""
        return self.read_tag(self.tag_map.images) is not None



"""
Implements all functionality pertaining to writing and deleting metadata/tags/properties for a :py:class:`LocalTrack`.
"""
from abc import ABCMeta, abstractmethod
from collections.abc import Mapping, Collection, Callable
from dataclasses import dataclass
from typing import Any

import mutagen

from musify.base import Result
from musify.libraries.core.object import Track
from musify.libraries.local.track._tags.base import TagProcessor
from musify.libraries.local.track.field import LocalTrackField as Tags
from musify.types import UnitIterable
from musify.utils import to_collection


@dataclass(frozen=True)
class SyncResultTrack(Result):
    """Stores the results of a sync with local track"""
    #: Were changes to the file on the disk made.
    saved: bool
    #: Map of the tag updated and the index of the condition it satisfied to be updated.
    updated: Mapping[Tags, int]


class TagWriter[T: mutagen.FileType](TagProcessor, metaclass=ABCMeta):
    """Functionality for updating and removing tags/metadata/properties from a mutagen object."""

    __slots__ = ()

    #: The date format to use when saving string representations of dates to tag values
    date_format = "%Y-%m-%d"

    def delete_tags(self, tags: UnitIterable[Tags] = (), dry_run: bool = True) -> SyncResultTrack:
        """
        Remove tags from file.

        :param tags: Tags to remove.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: List of tags that have been removed.
        """
        if tags is None or (isinstance(tags, Collection) and len(tags) == 0):
            return SyncResultTrack(saved=False, updated={})

        tag_names = set(Tags.to_tags(tags))
        removed = set()
        for tag_name in tag_names:
            if self._clear_tag(tag_name, dry_run):
                removed.update(Tags.from_name(tag_name))

        save = not dry_run and len(removed) > 0
        if save:
            self.file.save()

        removed = sorted(removed, key=lambda x: Tags.all().index(x))
        return SyncResultTrack(saved=save, updated={u: 0 for u in removed})

    def _clear_tag(self, tag_name: str, dry_run: bool = True) -> bool:
        """
        Remove a tag by its tag name from the loaded file object in memory.

        :param tag_name: Tag name as found in :py:class:`TagMap` to remove.
        :param dry_run: Run function, but do not modify the loaded file in memory.
        :return: True if tag has been remove, False otherwise.
        """
        removed = False

        tag_ids = self.tag_map[tag_name]
        if tag_ids is None or len(tag_ids) is None:
            return removed

        for tag_id in tag_ids:
            if tag_id in self.file and self.file[tag_id]:
                if not dry_run:
                    del self.file[tag_id]
                removed = True

        return removed

    def clear_loaded_images(self) -> bool:
        """
        Clear the loaded embedded images for this track.
        Does not alter the actual file in any way, only the loaded object in memory.
        """
        tag_names = Tags.IMAGES.to_tag()
        removed = False
        for tag_name in tag_names:
            removed = removed or self._clear_tag(tag_name, dry_run=False)

        return removed

    def write(
            self,
            source: Track,
            target: Track,
            tags: UnitIterable[Tags] = Tags.ALL,
            replace: bool = False,
            dry_run: bool = True
    ):
        """
        Write the tags from the ``target`` track to the ``source`` track.
        Filter the tags written by supplying ``tags``.

        :param source: The source track i.e. the track object representing the currently saved file on the drive.
        :param target: The target track i.e. the track object containing new tags with which to update the file.
        :param tags: The tags to be updated.
        :param replace: Destructively overwrite the tag on the file if the ``source`` and ``target`` tags differ.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The index number of the conditional that was met to warrant updating the file's tags.
            None if none of the conditions were met.
        """
        tags: set[Tags] = to_collection(tags, set)
        if Tags.ALL in tags:
            tags = set(Tags.all(only_tags=True))

        if any(f in tags for f in {Tags.TRACK, Tags.TRACK_NUMBER, Tags.TRACK_TOTAL}):
            tags -= {Tags.TRACK_NUMBER, Tags.TRACK_TOTAL}
            tags.add(Tags.TRACK)

        if any(f in tags for f in {Tags.DATE, Tags.YEAR, Tags.MONTH, Tags.DAY}):
            tags -= {Tags.YEAR, Tags.MONTH, Tags.DAY}
            tags.add(Tags.DATE)

        if any(f in tags for f in {Tags.DISC, Tags.DISC_NUMBER, Tags.DISC_TOTAL}):
            tags -= {Tags.DISC_NUMBER, Tags.DISC_TOTAL}
            tags.add(Tags.DISC)

        updated = {}
        for tag in tags:
            method: Callable[Any, Mapping[Tags, int] | int | None] = getattr(self, f"write_{tag.name.lower()}")
            result = method(source=source, target=target, replace=replace, dry_run=dry_run)
            if result is None:
                continue

            if isinstance(result, Mapping):
                updated |= result
            else:
                updated[tag] = result

        save = not dry_run and len(updated) > 0
        if save:
            self.file.save()

        return SyncResultTrack(saved=save, updated=updated)

    def write_tag(self, tag_id: str | None, tag_value: Any, dry_run: bool = True) -> bool | None:
        """
        Generic method for updating a tag value in the file.

        :param tag_id: ID of the tag for this file type.
        :param tag_value: New value to assign.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The index number of the conditional that was met to warrant updating the file's tags.
            None if none of the conditions were met.
        """
        if tag_id is None:
            return False

        if tag_value is None:
            if not dry_run and tag_id in self.file and self.file[tag_id] is not None:
                del self.file[tag_id]
                return True
            return False

        return self._write_tag(tag_id=tag_id, tag_value=tag_value, dry_run=dry_run)

    @abstractmethod
    def _write_tag(self, tag_id: str | None, tag_value: Any, dry_run: bool = True) -> bool:
        """Implementation of tag writer specific to this file type."""
        raise NotImplementedError

    def write_title(self, source: Track, target: Track, replace: bool = False, dry_run: bool = True) -> int | None:
        """
        Write track title tags to file if appropriate related conditions are met.

        :param source: The source track i.e. the track object representing the currently saved file on the drive.
        :param target: The target track i.e. the track object containing new tags with which to update the file.
        :param replace: Destructively overwrite the tag on the file if the ``source`` and ``target`` tags differ.
        :param replace: Destructively overwrite the tag on the file if the ``source`` and ``target`` tags differ.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The index number of the conditional that was met to warrant updating the file's tags.
            None if none of the conditions were met.
        """
        conditionals = [
            source.title is None and target.title is not None,
            replace and source.title != target.title
        ]
        if any(conditionals) and self._write_title(track=target, dry_run=dry_run):
            return next(i for i, c in enumerate(conditionals) if c)

    def _write_title(self, track: Track, dry_run: bool = True) -> bool:
        """
        Write the title tag from ``track`` to the stored file.

        :param track: The track with the tag to be written.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: True if the file has been updated or would have been written when ``dry_run`` is True.
        """
        return self.write_tag(next(iter(self.tag_map.title), None), track.title, dry_run)

    def write_artist(self, source: Track, target: Track, replace: bool = False, dry_run: bool = True) -> int | None:
        """
        Write the track artist tags to file if appropriate related conditions are met.

        :param source: The source track i.e. the track object representing the currently saved file on the drive.
        :param target: The target track i.e. the track object containing new tags with which to update the file.
        :param replace: Destructively overwrite the tag on the file if the ``source`` and ``target`` tags differ.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The index number of the conditional that was met to warrant updating the file's tags.
            None if none of the conditions were met.
        """
        conditionals = [
            source.artist is None and target.artist is not None,
            replace and source.artist != target.artist
        ]
        if any(conditionals) and self._write_artist(track=target, dry_run=dry_run):
            return next(i for i, c in enumerate(conditionals) if c)

    def _write_artist(self, track: Track, dry_run: bool = True) -> bool:
        """
        Write the artist tag from ``track`` to the stored file.

        :param track: The track with the tag to be written.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: True if the file has been updated or would have been written when ``dry_run`` is True.
        """
        return self.write_tag(next(iter(self.tag_map.artist), None), track.artist, dry_run)

    def write_album(self, source: Track, target: Track, replace: bool = False, dry_run: bool = True) -> int | None:
        """
        Write the track album tags to file if appropriate related conditions are met.

        :param source: The source track i.e. the track object representing the currently saved file on the drive.
        :param target: The target track i.e. the track object containing new tags with which to update the file.
        :param replace: Destructively overwrite the tag on the file if the ``source`` and ``target`` tags differ.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The index number of the conditional that was met to warrant updating the file's tags.
            None if none of the conditions were met.
        """
        conditionals = [
            source.album is None and target.album is not None,
            replace and source.album != target.album
        ]
        if any(conditionals) and self._write_album(track=target, dry_run=dry_run):
            return next(i for i, c in enumerate(conditionals) if c)

    def _write_album(self, track: Track, dry_run: bool = True) -> bool:
        """
        Write the album tag from ``track`` to the stored file.

        :param track: The track with the tag to be written.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: True if the file has been updated or would have been written when ``dry_run`` is True.
        """
        return self.write_tag(next(iter(self.tag_map.album), None), track.album, dry_run)

    def write_album_artist(
            self, source: Track, target: Track, replace: bool = False, dry_run: bool = True
    ) -> int | None:
        """
        Write the track album artist tags to file if appropriate related conditions are met.

        :param source: The source track i.e. the track object representing the currently saved file on the drive.
        :param target: The target track i.e. the track object containing new tags with which to update the file.
        :param replace: Destructively overwrite the tag on the file if the ``source`` and ``target`` tags differ.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The index number of the conditional that was met to warrant updating the file's tags.
            None if none of the conditions were met.
        """
        conditionals = [
            source.album_artist is None and target.album_artist is not None,
            replace and source.album_artist != target.album_artist
        ]
        if any(conditionals) and self._write_album_artist(track=target, dry_run=dry_run):
            return next(i for i, c in enumerate(conditionals) if c)

    def _write_album_artist(self, track: Track, dry_run: bool = True) -> bool:
        """
        Write the album artist tag from ``track`` to the stored file.

        :param track: The track with the tag to be written.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: True if the file has been updated or would have been written when ``dry_run`` is True.
        """
        return self.write_tag(next(iter(self.tag_map.album_artist), None), track.album_artist, dry_run)

    def write_track(self, source: Track, target: Track, replace: bool = False, dry_run: bool = True) -> int | None:
        """
        Write the track number and track total tags to file if appropriate related conditions are met.

        :param source: The source track i.e. the track object representing the currently saved file on the drive.
        :param target: The target track i.e. the track object containing new tags with which to update the file.
        :param replace: Destructively overwrite the tag on the file if the ``source`` and ``target`` tags differ.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The index number of the conditional that was met to warrant updating the file's tags.
            None if none of the conditions were met.
        """
        conditionals = [
            source.track_number is None and source.track_total is None and
            (target.track_number is not None or target.track_total is not None),
            replace and (source.track_number != target.track_number or source.track_total != target.track_total)
        ]
        if any(conditionals) and self._write_track(track=target, dry_run=dry_run):
            return next(i for i, c in enumerate(conditionals) if c)

    def _write_track(self, track: Track, dry_run: bool = True) -> bool:
        """
        Write the track number and track total tags from ``track`` to the stored file.

        :param track: The track with the tag to be written.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: True if the file has been updated or would have been written when ``dry_run`` is True.
        """
        tag_id_number = next(iter(self.tag_map.track_number), None)
        tag_id_total = next(iter(self.tag_map.track_total), None)

        if tag_id_number != tag_id_total and track.track_total is not None:
            number_updated = self.write_tag(tag_id_number, str(track.track_number).zfill(2), dry_run)
            total_updated = self.write_tag(tag_id_total, str(track.track_total).zfill(2), dry_run)
            return number_updated or total_updated
        elif track.track_total is not None:
            tag_value = self.num_sep.join([str(track.track_number).zfill(2), str(track.track_total).zfill(2)])
        else:
            tag_value = str(track.track_number).zfill(2)

        return self.write_tag(tag_id_number, tag_value, dry_run)

    def write_genres(self, source: Track, target: Track, replace: bool = False, dry_run: bool = True) -> int | None:
        """
        Write the track genre tags to file if appropriate related conditions are met.

        :param source: The source track i.e. the track object representing the currently saved file on the drive.
        :param target: The target track i.e. the track object containing new tags with which to update the file.
        :param replace: Destructively overwrite the tag on the file if the ``source`` and ``target`` tags differ.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The index number of the conditional that was met to warrant updating the file's tags.
            None if none of the conditions were met.
        """
        conditionals = [source.genres is None and bool(target.genres), replace and source.genres != target.genres]
        if any(conditionals) and self._write_genres(track=target, dry_run=dry_run):
            return next(i for i, c in enumerate(conditionals) if c)

    def _write_genres(self, track: Track, dry_run: bool = True) -> bool:
        """
        Write the genre tags from ``track`` to the stored file.

        :param track: The track with the tag to be written.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: True if the file has been updated or would have been written when ``dry_run`` is True.
        """
        return self.write_tag(next(iter(self.tag_map.genres), None), track.genres, dry_run)

    def write_date(
            self, source: Track, target: Track, replace: bool = False, dry_run: bool = True
    ) -> dict[Tags, int] | None:
        """
        Write the track date and/or year/month/day tags to file if appropriate related conditions are met.

        :param source: The source track i.e. the track object representing the currently saved file on the drive.
        :param target: The target track i.e. the track object containing new tags with which to update the file.
        :param replace: Destructively overwrite the tag on the file if the ``source`` and ``target`` tags differ.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The index number of the conditional that was met to warrant updating the file's tags.
            None if none of the conditions were met.
        """
        # date is just a composite of year + month + day, can safely ignore this in the conditionals
        values_exist = any({target.year is not None, target.month is not None, target.day is not None})
        conditionals = [
            source.year is None and source.month is None and source.day is None and values_exist,
            replace and (source.year != target.year or source.month != target.month or source.day != target.day)
        ]

        if not any(conditionals):
            return

        date, year, month, day = self._write_date(track=target, dry_run=dry_run)

        updated = {}
        condition = next(i for i, c in enumerate(conditionals) if c)
        if date:
            updated[Tags.DATE] = condition
        if year:
            updated[Tags.YEAR] = condition
        if month:
            updated[Tags.MONTH] = condition
        if day:
            updated[Tags.DAY] = condition

        return updated if updated else None

    def _write_date(self, track: Track, dry_run: bool = True) -> tuple[bool, bool, bool, bool]:
        """
        Write the date, year, month, and/or day tags from ``track`` to the stored file.

        :param track: The track with the tag to be written.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: True for each updated tag in the order (date, year, month, day)
            if the file has been updated or would have been written when ``dry_run`` is True.
        """
        date_str = track.date.strftime(self.date_format) if track.date else None
        if not date_str:
            date_str = f"{track.year}-{str(track.month).zfill(2)}" if track.month else str(track.year)
        date = self.write_tag(next(iter(self.tag_map.date), None), date_str, dry_run)
        if date:
            return True, False, False, False

        year = self.write_tag(next(iter(self.tag_map.year), None), track.year, dry_run)
        month = self.write_tag(next(iter(self.tag_map.month), None), track.month, dry_run)
        day = self.write_tag(next(iter(self.tag_map.day), None), track.day, dry_run)

        return date, year, month, day

    def write_bpm(self, source: Track, target: Track, replace: bool = False, dry_run: bool = True) -> int | None:
        """
        Write track bpm tags to file if appropriate related conditions are met.

        :param source: The source track i.e. the track object representing the currently saved file on the drive.
        :param target: The target track i.e. the track object containing new tags with which to update the file.
        :param replace: Destructively overwrite the tag on the file if the ``source`` and ``target`` tags differ.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The index number of the conditional that was met to warrant updating the file's tags.
            None if none of the conditions were met.
        """
        source_bpm = int(source.bpm if source.bpm is not None else 0)
        target_bpm = int(target.bpm if target.bpm is not None else 0)

        conditionals = [
            source.bpm is None and target.bpm is not None and target_bpm > 30,
            source_bpm < 30 < target_bpm,
            replace and source_bpm != target_bpm
        ]

        if any(conditionals) and self._write_bpm(track=target, dry_run=dry_run):
            return next(i for i, c in enumerate(conditionals) if c)

    def _write_bpm(self, track: Track, dry_run: bool = True) -> bool:
        """
        Write the bpm tag from ``track`` to the stored file.

        :param track: The track with the tag to be written.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: True if the file has been updated or would have been written when ``dry_run`` is True.
        """
        return self.write_tag(next(iter(self.tag_map.bpm), None), track.bpm, dry_run)

    def write_key(self, source: Track, target: Track, replace: bool = False, dry_run: bool = True) -> int | None:
        """
        Write track key tags to file if appropriate related conditions are met.

        :param source: The source track i.e. the track object representing the currently saved file on the drive.
        :param target: The target track i.e. the track object containing new tags with which to update the file.
        :param replace: Destructively overwrite the tag on the file if the ``source`` and ``target`` tags differ.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The index number of the conditional that was met to warrant updating the file's tags.
            None if none of the conditions were met.
        """
        conditionals = [source.key is None and target.key is not None, replace and source.key != target.key]

        if any(conditionals) and self._write_key(track=target, dry_run=dry_run):
            return next(i for i, c in enumerate(conditionals) if c)

    def _write_key(self, track: Track, dry_run: bool = True) -> bool:
        """
        Write the key tag from ``track`` to the stored file.

        :param track: The track with the tag to be written.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: True if the file has been updated or would have been written when ``dry_run`` is True.
        """
        return self.write_tag(next(iter(self.tag_map.key), None), track.key, dry_run)

    def write_disc(self, source: Track, target: Track, replace: bool = False, dry_run: bool = True) -> int | None:
        """
        Write track dic number and disc total tags to file if appropriate related conditions are met.

        :param source: The source track i.e. the track object representing the currently saved file on the drive.
        :param target: The target track i.e. the track object containing new tags with which to update the file.
        :param replace: Destructively overwrite the tag on the file if the ``source`` and ``target`` tags differ.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The index number of the conditional that was met to warrant updating the file's tags.
            None if none of the conditions were met.
        """
        conditionals = [
            source.disc_number is None and source.disc_total is None and
            (target.disc_number is not None or target.disc_total is not None),
            replace and (source.disc_number != target.disc_number or source.disc_total != target.disc_total)
        ]
        if any(conditionals) and self._write_disc(track=target, dry_run=dry_run):
            return next(i for i, c in enumerate(conditionals) if c)

    def _write_disc(self, track: Track, dry_run: bool = True) -> bool:
        """
        Write the disc number and disc total tags from ``track`` to the stored file.

        :param track: The track with the tag to be written.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: True if the file has been updated or would have been written when ``dry_run`` is True.
        """
        tag_id_number = next(iter(self.tag_map.disc_number), None)
        tag_id_total = next(iter(self.tag_map.disc_total), None)
        fill = len(str(track.disc_total)) if track.disc_total is not None else 1

        if tag_id_number != tag_id_total and track.disc_total is not None:
            number_updated = self.write_tag(tag_id_number, str(track.disc_number).zfill(fill), dry_run)
            total_updated = self.write_tag(tag_id_total, str(track.disc_total).zfill(fill), dry_run)
            return number_updated or total_updated
        elif track.disc_total is not None:
            tag_value = self.num_sep.join([str(track.disc_number).zfill(fill), str(track.disc_total).zfill(fill)])
        else:
            tag_value = str(track.disc_number).zfill(fill)

        return self.write_tag(tag_id_number, tag_value, dry_run)

    def write_compilation(
            self, source: Track, target: Track, replace: bool = False, dry_run: bool = True
    ) -> int | None:
        """
        Write track compilation tags to file if appropriate related conditions are met.

        :param source: The source track i.e. the track object representing the currently saved file on the drive.
        :param target: The target track i.e. the track object containing new tags with which to update the file.
        :param replace: Destructively overwrite the tag on the file if the ``source`` and ``target`` tags differ.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The index number of the conditional that was met to warrant updating the file's tags.
            None if none of the conditions were met.
        """
        conditionals = [
            source.compilation is None and target.compilation is not None,
            replace and source.compilation != target.compilation
        ]
        if any(conditionals) and self._write_compilation(track=target, dry_run=dry_run):
            return next(i for i, c in enumerate(conditionals) if c)

    def _write_compilation(self, track: Track, dry_run: bool = True) -> bool:
        """
        Write the compilation tag from ``track`` to the stored file.

        :param track: The track with the tag to be written.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: True if the file has been updated or would have been written when ``dry_run`` is True.
        """
        return self.write_tag(next(iter(self.tag_map.compilation), None), int(track.compilation), dry_run)

    def write_comments(self, source: Track, target: Track, replace: bool = False, dry_run: bool = True) -> int | None:
        """
        Write track comments tags to file if appropriate related conditions are met.

        :param source: The source track i.e. the track object representing the currently saved file on the drive.
        :param target: The target track i.e. the track object containing new tags with which to update the file.
        :param replace: Destructively overwrite the tag on the file if the ``source`` and ``target`` tags differ.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The index number of the conditional that was met to warrant updating the file's tags.
            None if none of the conditions were met.
        """
        conditionals = [
            source.comments is None and bool(target.comments),
            replace and source.comments != target.comments
        ]
        if any(conditionals) and self._write_comments(track=target, dry_run=dry_run):
            return next(i for i, c in enumerate(conditionals) if c)

    def _write_comments(self, track: Track, dry_run: bool = True) -> bool:
        """
        Write the comments tags from ``track`` to the stored file.

        :param track: The track with the tag to be written.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: True if the file has been updated or would have been written when ``dry_run`` is True.
        """
        return self.write_tag(next(iter(self.tag_map.comments), None), track.comments, dry_run)

    # noinspection PyUnusedLocal
    def write_uri(self, source: Track, target: Track, replace: bool = False, dry_run: bool = True) -> int | None:
        """
        Write track URI tag to file if appropriate related conditions are met.

        :param source: The source track i.e. the track object representing the currently saved file on the drive.
        :param target: The target track i.e. the track object containing new tags with which to update the file.
        :param replace: Ignored.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The index number of the conditional that was met to warrant updating the file's tags.
            None if none of the conditions were met.
        """
        if not self.remote_wrangler:
            return

        conditionals = [source.uri != target.uri or source.has_uri != target.has_uri]

        if any(conditionals) and self._write_uri(track=target, dry_run=dry_run):
            return next(i for i, c in enumerate(conditionals) if c)

    def _write_uri(self, track: Track, dry_run: bool = True) -> bool:
        """
        Write the URI tag from ``track`` to the stored file.

        :param track: The track with the tag to be written.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: True if the file has been updated or would have been written when ``dry_run`` is True.
        """
        tag_id = next(iter(self.tag_map[self.uri_tag.name.lower()]), None)
        tag_value = self.remote_wrangler.unavailable_uri_dummy if not track.has_uri else track.uri
        return self.write_tag(tag_id, tag_value, dry_run)

    def write_images(self, source: Track, target: Track, replace: bool = False, dry_run: bool = True) -> int | None:
        """
        Write images to file if appropriate related conditions are met.

        :param source: The source track i.e. the track object representing the currently saved file on the drive.
        :param target: The target track i.e. the track object containing new tags with which to update the file.
        :param replace: Destructively overwrite the tag on the file if the ``source`` and ``target`` tags differ.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: The index number of the conditional that was met to warrant updating the file's tags.
            None if none of the conditions were met.
        """
        conditionals = [source.has_image is False and bool(target.image_links), replace and bool(target.image_links)]

        if any(conditionals) and self._write_images(track=target, dry_run=dry_run):
            return next(i for i, c in enumerate(conditionals) if c)

    @abstractmethod
    def _write_images(self, track: Track, dry_run: bool = True) -> bool:
        """
        Write the images from the ``image_links`` in ``track`` to the stored file.

        :param track: The track with the tag to be written.
        :param dry_run: Run function, but do not modify the file on the disk.
        :return: True if the file has been updated or would have been written when ``dry_run`` is True.
        """
        raise NotImplementedError



"""
Processor that limits the items in a given collection of items
"""
from collections.abc import Collection, MutableSequence
from functools import reduce
from operator import mul
from random import shuffle

from musify.base import MusifyItem, HasLength
from musify.field import Fields
from musify.file.base import File
from musify.libraries.core.object import Track
from musify.processors.base import DynamicProcessor, dynamicprocessormethod
from musify.processors.exception import LimiterProcessorError
from musify.processors.sort import ItemSorter
from musify.types import MusifyEnum


class LimitType(MusifyEnum):
    """Represents the possible limit types to apply when filtering a playlist."""
    ITEMS = 0
    ALBUMS = 1

    SECONDS = 11
    MINUTES = 12
    HOURS = 13
    DAYS = 14
    WEEKS = 15

    BYTES = 20
    KILOBYTES = 21
    MEGABYTES = 22
    GIGABYTES = 23
    TERABYTES = 24


class ItemLimiter(DynamicProcessor):
    """
    Sort items in-place based on given conditions.

    :param limit: The number of items to limit to. A value of 0 applies no limiting.
    :param on: The type to limit on e.g. items, albums, minutes.
    :param sorted_by: When limiting, sort the collection of items by this function first.
    :param allowance: When limiting on bytes or length, add this extra allowance factor to
        the max size limit on comparison.
        e.g. say the limiter currently has 29 minutes worth of songs in its final list and the max limit is 30 minutes.
        The limiter has to now consider whether to include the next song it sees with length 3 minutes.
        With an allowance of 0, this song will not be added.
        However, with an allowance of say 1.33 it will as the max limit for this comparison becomes 30 * 1.33 = 40.
        Now, with 32 minutes worth of songs in the final playlist, the limit is >30 minutes and the limiter stops
        processing.
    """

    __slots__ = ("limit_max", "kind", "allowance")

    @classmethod
    def _processor_method_fmt(cls, name: str) -> str:
        return "_" + cls._pascal_to_snake(name)

    @property
    def limit_sort(self) -> str | None:
        """String representation of the sorting method to use before limiting"""
        return self._processor_name.lstrip("_")

    def __init__(
            self,
            limit: int = 0,
            on: LimitType = LimitType.ITEMS,
            sorted_by: str | None = None,
            allowance: float = 1.0,
    ):
        super().__init__()

        #: The number of items to limit to.
        self.limit_max = limit
        #: The type to limit on e.g. items, albums, minutes.
        self.kind = on
        #: When limiting on bytes or length, add this extra allowance factor to
        #: the max size limit on comparison.
        self.allowance = allowance

        self._set_processor_name(sorted_by, fail_on_empty=False)

    def __call__(self, *args, **kwargs) -> None:
        return self.limit(*args, **kwargs)

    def limit[T: MusifyItem](self, items: list[T], ignore: Collection[T] = ()) -> None:
        """
        Limit ``items`` in-place based on set conditions.

        :param items: The list of items to limit.
        :param ignore: list of items to ignore when limiting. i.e. keep them in the list regardless.
        """
        if len(items) == 0 or self.limit_max == 0:
            return

        if self._processor_name:  # sort the input items in-place if sort method given
            super().__call__(items)

        if ignore:  # filter out the ignore items if given
            items_limit = [item for item in items if item not in ignore]
            items_ignore = [item for item in items if item in ignore]
            items.clear()
            items.extend(items_ignore)
        else:  # make a copy of the given items and clear the original list
            items_limit = [t for t in items]
            items.clear()

        if self.kind == LimitType.ITEMS:
            items.extend(items_limit[:self.limit_max])
        elif self.kind == LimitType.ALBUMS:
            items.extend(self._limit_on_albums(items_limit))
        else:
            items.extend(self._limit_on_numeric(items_limit))

    def _limit_on_albums[T: MusifyItem](self, items: MutableSequence[T]) -> list[T]:
        seen_albums = []
        result = []

        for item in items:
            if not isinstance(item, Track):
                LimiterProcessorError("In order to limit on Album, all items must be of type 'Track'")

            if len(seen_albums) < self.limit_max and item.album not in seen_albums:
                # album limit not yet reached
                seen_albums.append(item.album)
            if item.album in seen_albums:
                result.append(item)

        return result

    def _limit_on_numeric[T: MusifyItem](self, items: MutableSequence[T]) -> list[T]:
        count = 0
        result = []

        for item in items:
            value = self._convert(item)
            if count + value <= self.limit_max * self.allowance:  # limit not yet reached
                result.append(item)
                count += value
            if count > self.limit_max:  # limit reached
                break

        return result

    def _convert(self, item: MusifyItem) -> float:
        """
        Convert units for item length or size and return the value.

        :raise ItemLimiterError: When the given limit type cannot be found
        """
        if 10 < self.kind.value < 20:
            if not isinstance(item, HasLength):
                raise LimiterProcessorError("The given item cannot be limited on length as it does not have a length.")

            factors = (1, 60, 60, 24, 7)[:self.kind.value % 10]
            return item.length / reduce(mul, factors, 1)

        elif 20 <= self.kind.value < 30:
            if not isinstance(item, File):
                raise LimiterProcessorError("The given item cannot be limited on bytes as it is not a file.")

            bytes_scale = 1000
            return item.size / (bytes_scale ** (self.kind.value % 10))

        else:
            raise LimiterProcessorError(f"Unrecognised LimitType: {self.kind}")

    @dynamicprocessormethod
    def _random(self, items: MutableSequence[MusifyItem]) -> None:
        shuffle(items)

    @dynamicprocessormethod
    def _highest_rating(self, items: list[MusifyItem]) -> None:
        ItemSorter.sort_by_field(items, Fields.RATING, reverse=True)

    @dynamicprocessormethod
    def _lowest_rating(self, items: list[MusifyItem]) -> None:
        ItemSorter.sort_by_field(items, Fields.RATING)

    @dynamicprocessormethod
    def _most_recently_played(self, items: list[MusifyItem]) -> None:
        ItemSorter.sort_by_field(items, Fields.LAST_PLAYED, reverse=True)

    @dynamicprocessormethod
    def _least_recently_played(self, items: list[MusifyItem]) -> None:
        ItemSorter.sort_by_field(items, Fields.LAST_PLAYED)

    @dynamicprocessormethod
    def _most_often_played(self, items: list[MusifyItem]) -> None:
        ItemSorter.sort_by_field(items, Fields.PLAY_COUNT, reverse=True)

    @dynamicprocessormethod
    def _least_often_played(self, items: list[MusifyItem]) -> None:
        ItemSorter.sort_by_field(items, Fields.PLAY_COUNT)

    @dynamicprocessormethod
    def _most_recently_added(self, items: list[MusifyItem]) -> None:
        ItemSorter.sort_by_field(items, Fields.DATE_ADDED, reverse=True)

    @dynamicprocessormethod
    def _least_recently_added(self, items: list[MusifyItem]) -> None:
        ItemSorter.sort_by_field(items, Fields.DATE_ADDED)

    def as_dict(self):
        return {
            "on": self.kind,
            "limit": self.limit_max,
            "sorted_by": self.limit_sort,
            "allowance": self.allowance
        }



"""
Processors that filter down objects and data types based on some given configuration.
"""
from __future__ import annotations

from collections.abc import Collection, Sequence, Mapping
from pathlib import Path
from typing import Any, Self

from musify.base import MusifyObject
from musify.processors.base import Filter, FilterComposite
from musify.processors.compare import Comparer
from musify.types import UnitCollection


class FilterDefinedList[T: str | Path | MusifyObject](Filter[T], Collection[T]):

    __slots__ = ("values",)

    @property
    def ready(self):
        return len(self.values) > 0

    def __init__(self, values: Collection[T] = (), *_, **__):
        super().__init__()
        #: The values to include when processing for this filter
        self.values: Collection[T] = values

    def __call__(self, *args, **kwargs) -> Collection[T]:
        return self.process(*args, **kwargs)

    def process(self, values: Collection[T] | None = None, *_, **__) -> Collection[T]:
        """Returns all ``values`` that match this filter's settings"""
        if self.ready:
            matches = [value for value in values if self.transform(value) in self.values]
            if isinstance(self.values, Sequence):
                matches = sorted((self.values.index(self.transform(match)), match) for match in matches)
                return [match[1] for match in matches]
            return matches
        return values

    def as_dict(self) -> dict[str, Any]:
        return {"values": self.values}

    def __iter__(self):
        return iter(self.values)

    def __len__(self):
        return len(self.values)

    def __contains__(self, item: Any):
        return item in self.values


class FilterComparers[T: str | MusifyObject](Filter[T]):

    __slots__ = ("comparers", "match_all")

    @property
    def ready(self):
        return len(self.comparers) > 0

    def __init__(
            self,
            comparers: UnitCollection[Comparer] | Mapping[Comparer, tuple[bool, Self]] = (),
            match_all: bool = True,
            *_,
            **__
    ):
        super().__init__()
        if isinstance(comparers, Comparer):
            comparers = [comparers]
        if not isinstance(comparers, Mapping):
            comparers = {comparer: (False, FilterComparers()) for comparer in comparers}

        #: The comparers to use when processing for this filter
        #: When a mapping with other :py:class:`FilterComparers` is given,
        #: will apply this sub-filter to all parent filters when processing
        self.comparers: Mapping[Comparer, tuple[bool, Self]] = comparers
        #: When true, only include those items that match on all comparers
        self.match_all: bool = match_all

    def __call__(self, *args, **kwargs) -> Collection[T]:
        return self.process(*args, **kwargs)

    def process(self, values: Collection[T], reference: T | None = None, *_, **__) -> Collection[T]:
        if not self.ready:
            return values

        def run_comparer(c: Comparer, v: T) -> bool:
            """Run the comparer ``c`` for the given value ``v``"""
            return c(self.transform(v), reference=reference) if c.expected is None else c(self.transform(v))

        def run_sub_filter(s: FilterComparers, c: bool, v: list[T]) -> Collection[T]:
            """Run the sub_filter ``s`` for the given value ``v`` and return the results"""
            if not s.ready:
                return v
            elif c:
                return sub_filter(v, reference=reference)
            else:
                v.extend([value for value in sub_filter(values, reference=reference) if value not in v])
                return v

        if self.match_all:
            for comparer, (combine, sub_filter) in self.comparers.items():
                matched = [value for value in values if run_comparer(comparer, value)]
                values = run_sub_filter(sub_filter, combine, matched)
            return values

        matches = []
        for comparer, (combine, sub_filter) in self.comparers.items():
            matched = [value for value in values if value not in matches and run_comparer(comparer, value)]
            matched = run_sub_filter(sub_filter, combine, matched)
            matches.extend([value for value in matched if value not in matches])

        return matches

    def as_dict(self) -> dict[str, Any]:
        return {"comparers": self.comparers, "match_all": self.match_all}


###########################################################################
## Composites
###########################################################################
class FilterIncludeExclude[T: Any, U: Filter, V: Filter](FilterComposite[T]):

    __slots__ = ("include", "exclude")

    def __init__(self, include: U, exclude: V, *_, **__):
        super().__init__(include, exclude)
        #: The filter that, when processed, returns items to include
        self.include: U = include
        #: The filter that, when processed, returns items to exclude
        self.exclude: V = exclude

    def __call__(self, *args, **kwargs) -> list[T]:
        return self.process(*args, **kwargs)

    def process(self, values: Collection[T], *_, **__) -> list[T]:
        """Filter down ``values`` that match this filter's settings from"""
        values = self.include.process(values) if self.include.ready else values
        exclude = self.exclude.process(values) if self.exclude.ready else ()
        return [v for v in values if v not in exclude]

    def as_dict(self) -> dict[str, Any]:
        return {"include": self.include, "exclude": self.exclude}



"""
The WMA implementation of a :py:class:`LocalTrack`.
"""
import struct
from collections.abc import Collection, Iterable
from io import BytesIO
from typing import Any

import mutagen
import mutagen.asf
import mutagen.id3

from musify.field import TagMap
from musify.file.image import open_image, get_image_bytes
# noinspection PyProtectedMember
from musify.libraries.local.track._tags import TagReader, TagWriter
from musify.libraries.local.track.track import LocalTrack

try:
    from PIL import Image, UnidentifiedImageError
except ImportError:
    Image = None
    UnidentifiedImageError = None


class _WMATagReader(TagReader[mutagen.asf.ASF]):

    __slots__ = ()

    def read_tag(self, tag_ids: Iterable[str]) -> list[Any] | None:
        # WMA tag values are returned as mutagen.asf._attrs.ASFUnicodeAttribute
        values = []
        for tag_id in tag_ids:
            value: Collection[mutagen.asf.ASFBaseAttribute] = self.file.get(tag_id)
            if value is None:
                # skip null or empty/blank strings
                continue

            values.extend([v.value for v in value if not (isinstance(value, str) and len(value.strip()) == 0)])

        return values if len(values) > 0 else None

    def read_images(self):
        if Image is None:
            return

        values = self.read_tag(self.tag_map.images)
        if values is None:
            return

        images: list[Image.Image] = []
        for i, value in enumerate(values):
            try:
                # first attempt to open image from bytes; assumes bytes value refer only to the image data
                images.append(Image.open(BytesIO(value)))
                continue
            except UnidentifiedImageError:
                # the bytes are encoded per WMA spec
                # bytes need to be analysed first to extract the bytes that refer to the image data
                pass

            v_type, v_size = struct.unpack_from(b"<bi", value)

            pos = 5
            mime = b""
            while value[pos:pos + 2] != b"\x00\x00":
                mime += value[pos:pos + 2]
                pos += 2

            pos += 2
            description = b""

            while value[pos:pos + 2] != b"\x00\x00":
                description += value[pos:pos + 2]
                pos += 2
            pos += 2

            image_data: bytes = value[pos:pos + v_size]
            images.append(Image.open(BytesIO(image_data)))

        return images


class _WMATagWriter(TagWriter[mutagen.asf.ASF]):

    __slots__ = ()

    def _write_tag(self, tag_id: str | None, tag_value: Any, dry_run: bool = True) -> bool:
        if not dry_run:
            if isinstance(tag_value, (list, set, tuple)):
                if all(isinstance(v, mutagen.asf.ASFByteArrayAttribute) for v in tag_value):
                    self.file[tag_id] = tag_value
                else:
                    self.file[tag_id] = [mutagen.asf.ASFUnicodeAttribute(str(v)) for v in tag_value]
            else:
                self.file[tag_id] = mutagen.asf.ASFUnicodeAttribute(str(tag_value))
        return True

    def _write_images(self, track: LocalTrack, dry_run: bool = True) -> bool:
        tag_id = next(iter(self.tag_map.images), None)

        updated = False
        tag_value = []
        for image_kind, image_link in track.image_links.items():
            image_kind_attr = image_kind.upper().replace(" ", "_")
            image_type: mutagen.id3.PictureType = getattr(mutagen.id3.PictureType, image_kind_attr)

            image = open_image(image_link)
            data = get_image_bytes(image)
            tag_data = struct.pack("<bi", image_type, len(data))
            tag_data += Image.MIME[image.format].encode("utf-16") + b"\x00\x00"  # mime
            tag_data += "".encode("utf-16") + b"\x00\x00"  # description
            tag_data += data

            tag_value.append(mutagen.asf.ASFByteArrayAttribute(tag_data))
            image.close()

        if len(tag_value) > 0:
            updated = self.write_tag(tag_id, tag_value, dry_run)

        track.has_image = updated or track.has_image
        return updated


class WMA(LocalTrack[mutagen.asf.ASF, _WMATagReader, _WMATagWriter]):

    __slots__ = ()

    valid_extensions = frozenset({".wma"})

    #: Map of human-friendly tag name to ID3 tag ids for a given file type
    tag_map = TagMap(
        title=["Title"],
        artist=["Author"],
        album=["WM/AlbumTitle"],
        album_artist=["WM/AlbumArtist"],
        track_number=["WM/TrackNumber"],
        track_total=["TotalTracks", "WM/TrackNumber"],
        genres=["WM/Genre"],
        year=["WM/Year", "WM/OriginalReleaseYear"],
        bpm=["WM/BeatsPerMinute"],
        key=["WM/InitialKey"],
        disc_number=["WM/PartOfSet"],
        disc_total=["WM/PartOfSet"],
        compilation=["COMPILATION"],
        comments=["Description", "WM/Comments"],
        images=["WM/Picture"],
    )

    def _create_reader(self, file: mutagen.asf.ASF):
        return _WMATagReader(file, tag_map=self.tag_map, remote_wrangler=self._remote_wrangler)

    def _create_writer(self, file: mutagen.asf.ASF):
        return _WMATagWriter(file, tag_map=self.tag_map, remote_wrangler=self._remote_wrangler)



"""
The M4A implementation of a :py:class:`LocalTrack`.
"""
from collections.abc import Iterable
from io import BytesIO
from typing import Any

import mutagen
import mutagen.mp4

from musify.field import TagMap
from musify.file.image import open_image, get_image_bytes
# noinspection PyProtectedMember
from musify.libraries.local.track._tags import TagReader, TagWriter
from musify.libraries.local.track.track import LocalTrack
from musify.utils import to_collection

try:
    from PIL import Image
except ImportError:
    Image = None


class _M4ATagReader(TagReader[mutagen.mp4.MP4]):

    __slots__ = ()

    def read_tag(self, tag_ids: Iterable[str]) -> list[Any] | None:
        """Extract all tag values for a given list of tag IDs"""
        values = []
        for tag_id in tag_ids:
            value = self.file.get(tag_id)
            if value is None or (isinstance(value, str) and len(value.strip()) == 0):
                # skip null or empty/blank strings
                continue

            if tag_id.startswith("----:com.apple.iTunes") and isinstance(value, (list, set, tuple)):
                value = [v for val in value for v in val.decode("utf-8").split('\x00')]
            elif isinstance(value, bytes):
                value = value.decode("utf-8").split('\x00')

            values.extend(value) if isinstance(value, (list, set, tuple)) else values.append(value)

        return values if len(values) > 0 else None

    def read_track_number(self) -> int | None:
        values = self.read_tag(self.tag_map.track_number)
        return int(values[0][0]) if values is not None else None

    def read_track_total(self) -> int | None:
        values = self.read_tag(self.tag_map.track_total)
        return int(values[0][1]) if values is not None else None

    def read_key(self) -> str | None:
        values = self.read_tag(self.tag_map.key)
        return str(values[0][:]) if values is not None else None

    def read_disc_number(self) -> int | None:
        values = self.read_tag(self.tag_map.disc_number)
        return int(values[0][0]) if values is not None else None

    def read_disc_total(self) -> int | None:
        values = self.read_tag(self.tag_map.disc_total)
        return int(values[0][1]) if values is not None else None

    def read_images(self):
        if Image is None:
            return

        values = self.read_tag(self.tag_map.images)
        return [Image.open(BytesIO(bytes(value))) for value in values] if values is not None else None


class _M4ATagWriter(TagWriter[mutagen.mp4.MP4]):

    __slots__ = ()

    def _write_tag(self, tag_id: str | None, tag_value: Any, dry_run: bool = True) -> bool:
        if not dry_run:
            if tag_id.startswith("----:com.apple.iTunes"):
                self.file[tag_id] = [
                    mutagen.mp4.MP4FreeForm(str(v).encode("utf-8"), 1) for v in to_collection(tag_value)
                ]
            elif isinstance(tag_value, bool):
                self.file[tag_id] = tag_value
            elif isinstance(tag_value, tuple):
                self.file[tag_id] = [tag_value]
            else:
                self.file[tag_id] = to_collection(tag_value, list)
        return True

    def _write_track(self, track: LocalTrack, dry_run: bool = True) -> bool:
        tag_id = next(iter(self.tag_map.track_number), None)
        tag_value = (track.track_number, track.track_total)
        return self.write_tag(tag_id, tag_value, dry_run)

    def _write_date(self, track: LocalTrack, dry_run: bool = True) -> tuple[bool, bool, bool, bool]:
        date_str = track.date.strftime(self.date_format) if track.date else None
        if not date_str:
            date_str = f"{track.year}-{str(track.month).zfill(2)}" if track.month else str(track.year)
        date = self.write_tag(next(iter(self.tag_map.date), None), date_str, dry_run)
        if date:
            return date, False, False, False

        year = self.write_tag(next(iter(self.tag_map.year), None), str(track.year) if track.year else None, dry_run)
        month = self.write_tag(next(iter(self.tag_map.month), None), str(track.month) if track.month else None, dry_run)
        day = self.write_tag(next(iter(self.tag_map.day), None), str(track.day) if track.day else None, dry_run)

        return date, year, month, day

    def _write_bpm(self, track: LocalTrack, dry_run: bool = True) -> bool:
        bpm = int(track.bpm) if track.bpm is not None else None
        return self.write_tag(next(iter(self.tag_map.bpm), None), bpm, dry_run)

    def _write_disc(self, track: LocalTrack, dry_run: bool = True) -> bool:
        tag_id = next(iter(self.tag_map.disc_number), None)
        tag_value = (track.disc_number, track.disc_total)
        return self.write_tag(tag_id, tag_value, dry_run)

    def _write_compilation(self, track: LocalTrack, dry_run: bool = True) -> bool:
        return self.write_tag(next(iter(self.tag_map.compilation), None), track.compilation, dry_run)

    def _write_images(self, track: LocalTrack, dry_run: bool = True) -> bool:
        tag_id = next(iter(self.tag_map.images), None)

        updated = False
        tag_value = []
        for image_link in track.image_links.values():
            image = open_image(image_link)

            if image.format == "PNG":
                image_format = mutagen.mp4.MP4Cover.FORMAT_PNG
            else:
                image_format = mutagen.mp4.MP4Cover.FORMAT_JPEG

            tag_value.append(mutagen.mp4.MP4Cover(get_image_bytes(image), imageformat=image_format))
            image.close()

        if len(tag_value) > 0:
            updated = self.write_tag(tag_id, tag_value, dry_run)

        track.has_image = updated or track.has_image
        return updated


class M4A(LocalTrack[mutagen.mp4.MP4, _M4ATagReader, _M4ATagWriter]):

    __slots__ = ()

    valid_extensions = frozenset({".m4a"})

    # noinspection SpellCheckingInspection
    #: Map of human-friendly tag name to ID3 tag ids for a given file type
    tag_map = TagMap(
        title=["©nam"],
        artist=["©ART"],
        album=["©alb"],
        album_artist=["aART"],
        track_number=["trkn"],
        track_total=["trkn"],
        genres=["----:com.apple.iTunes:GENRE", "©gen", "gnre"],
        date=["©day"],
        bpm=["tmpo"],
        key=["----:com.apple.iTunes:INITIALKEY"],
        disc_number=["disk"],
        disc_total=["disk"],
        compilation=["cpil"],
        comments=["©cmt"],
        images=["covr"],
    )

    def _create_reader(self, file: mutagen.mp4.MP4):
        return _M4ATagReader(file, tag_map=self.tag_map, remote_wrangler=self._remote_wrangler)

    def _create_writer(self, file: mutagen.mp4.MP4):
        return _M4ATagWriter(file, tag_map=self.tag_map, remote_wrangler=self._remote_wrangler)
